# -*- coding: UTF-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 22.12.2023, 12:40:33

Purpose: RB: /tool/
"""

from typing import Dict, Optional, Any

from ....logstool.queue import LoggerQueue

from ....attribtool import ReadOnlyClass
from ....logstool.logs import LoggerClient
from ..base import BRouterOS, BDev
from ...network.connectors import IConnector


class _Elements(object, metaclass=ReadOnlyClass):
    """Keys definition class.

    For internal purpose only.
    """

    BANDWIDTH_SERVER: str = "bandwidth-server"
    CONNECTION: str = "connection"
    E_MAIL: str = "e-mail"
    GRAPHING: str = "graphing"
    HOST: str = "host"
    INBOX: str = "inbox"
    INTERFACE: str = "interface"
    LATENCY_DISTRIBUTION: str = "latency-distribution"
    MAC_SERVER: str = "mac-server"
    MAC_WINBOX: str = "mac-winbox"
    NETWATCH: str = "netwatch"
    PACKET: str = "packet"
    PACKET_TEMPLATE: str = "packet-template"
    PING: str = "ping"
    PORT: str = "port"
    PROTOCOL: str = "protocol"
    QUEUE: str = "queue"
    RAW: str = "raw"
    RAW_PACKET_TEMPLATE: str = "raw-packet-template"
    RESOURCE: str = "resource"
    ROMON: str = "romon"
    ROOT: str = "tool"
    SESSION: str = "session"
    SESSIONS: str = "sessions"
    SMS: str = "sms"
    SNIFFER: str = "sniffer"
    STATS: str = "stats"
    STREAM: str = "stream"
    TRAFFIC_GENERATOR: str = "traffic-generator"
    TRAFFIC_MONITOR: str = "traffic-monitor"


class RBTool(BRouterOS):
    """Tool class

    For command root: /tool/
    """

    def __init__(
        self,
        parent: BDev,
        connector: IConnector,
        qlog: Optional[LoggerQueue] = None,
        debug: bool = False,
        verbose: bool = False,
    ) -> None:
        """Constructor.

        ### Arguments:
        * parent: BDev - Parent device object.
        * connector: IConnector - Network connector interface for RouterOS communication.
        * qlog: Optional[LoggerQueue] - Logger queue for logging operations. Defaults to None.
        * debug: bool - Enable debug mode. Defaults to False.
        * verbose: bool - Enable verbose output. Defaults to False.
        """
        super().__init__(
            parent,
            connector,
            LoggerClient(queue=qlog, name=self._c_name),
            debug,
            verbose,
        )
        self.root = f"{_Elements.ROOT}/"

        # add elements
        elements: Dict[str, Any] = {
            _Elements.BANDWIDTH_SERVER: {_Elements.SESSION: {}},
            _Elements.E_MAIL: {},
            _Elements.GRAPHING: {
                _Elements.INTERFACE: {},
                _Elements.QUEUE: {},
                _Elements.RESOURCE: {},
            },
            _Elements.MAC_SERVER: {
                _Elements.MAC_WINBOX: {},
                _Elements.PING: {},
                _Elements.SESSIONS: {},
            },
            _Elements.NETWATCH: {},
            _Elements.ROMON: {_Elements.PORT: {}},
            _Elements.SMS: {_Elements.INBOX: {}},
            _Elements.SNIFFER: {
                _Elements.CONNECTION: {},
                _Elements.HOST: {},
                _Elements.PACKET: {},
                _Elements.PROTOCOL: {},
            },
            _Elements.TRAFFIC_GENERATOR: {
                _Elements.PACKET_TEMPLATE: {},
                _Elements.PORT: {},
                _Elements.RAW_PACKET_TEMPLATE: {},
                _Elements.STATS: {
                    _Elements.LATENCY_DISTRIBUTION: {},
                    _Elements.PORT: {},
                    _Elements.RAW: {},
                    _Elements.STREAM: {},
                },
                _Elements.STREAM: {},
            },
            _Elements.TRAFFIC_MONITOR: {},
        }

        # configure elements
        self._add_elements(self, elements)


# #[EOF]#######################################################################



"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 2024-05-12

Purpose: Provide shared helper types and mixins leveraged by device
abstractions.

Centralising reusable building blocks keeps device implementations focused on
vendor specific logic while importing from this package for common behaviour.
"""

# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 22.12.2023, 12:35:53

Purpose: RB: /ipv6/
"""

from typing import Dict, Optional, Any

from ....logstool.queue import LoggerQueue

from ....attribtool import ReadOnlyClass
from ....logstool.logs import LoggerClient
from ..base import BRouterOS, BDev
from ...network.connectors import IConnector


class _Elements(object, metaclass=ReadOnlyClass):
    """Keys definition class.

    For internal purpose only.
    """

    ADDRESS: str = "address"
    ADDRESS_LIST: str = "address-list"
    BINDING: str = "binding"
    CONNECTION: str = "connection"
    DEFAULT: str = "default"
    DHCP_CLIENT: str = "dhcp-client"
    DHCP_RELAY: str = "dhcp-relay"
    DHCP_SERVER: str = "dhcp-server"
    FILTER: str = "filter"
    FIREWALL: str = "firewall"
    MANGLE: str = "mangle"
    NAT: str = "nat"
    ND: str = "nd"
    NEIGHBOR: str = "neighbor"
    OPTION: str = "option"
    POOL: str = "pool"
    PREFIX: str = "prefix"
    RAW: str = "raw"
    ROOT: str = "ipv6"
    ROUTE: str = "route"
    SETS: str = "sets"
    SETTINGS: str = "settings"
    USED: str = "used"


class RBIpv6(BRouterOS):
    """IPv6 class

    For command root: /ipv6/
    """

    def __init__(
        self,
        parent: BDev,
        connector: IConnector,
        qlog: Optional[LoggerQueue] = None,
        debug: bool = False,
        verbose: bool = False,
    ) -> None:
        """Constructor.

        ### Arguments:
        * parent: BDev - Parent device object.
        * connector: IConnector - Network connector interface for RouterOS communication.
        * qlog: Optional[LoggerQueue] - Logger queue for logging operations. Defaults to None.
        * debug: bool - Enable debug mode. Defaults to False.
        * verbose: bool - Enable verbose output. Defaults to False.
        """
        super().__init__(
            parent,
            connector,
            LoggerClient(queue=qlog, name=self._c_name),
            debug,
            verbose,
        )
        self.root = f"{_Elements.ROOT}/"

        # add elements
        elements: Dict[str, Any] = {
            _Elements.ADDRESS: {},
            _Elements.DHCP_CLIENT: {_Elements.OPTION: {}},
            _Elements.DHCP_RELAY: {},
            _Elements.DHCP_SERVER: {
                _Elements.BINDING: {},
                _Elements.OPTION: {_Elements.SETS: {}},
            },
            _Elements.FIREWALL: {
                _Elements.ADDRESS_LIST: {},
                _Elements.CONNECTION: {},
                _Elements.FILTER: {},
                _Elements.MANGLE: {},
                _Elements.NAT: {},
                _Elements.RAW: {},
            },
            _Elements.ND: {_Elements.PREFIX: {_Elements.DEFAULT: {}}},
            _Elements.NEIGHBOR: {},
            _Elements.POOL: {_Elements.USED: {}},
            _Elements.ROUTE: {},
            _Elements.SETTINGS: {},
        }

        # configure elements
        self._add_elements(self, elements)


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 02.12.2023

Purpose: Sets of classes for various date/time operations.
"""

import calendar
from time import time
from datetime import datetime, timezone, timedelta
from typing import Optional, Tuple, Union
from inspect import currentframe

from .attribtool import NoNewAttributes
from .raisetool import Raise


class DateTime(NoNewAttributes):
    """A utility class for generating various datetime structures."""

    @classmethod
    def now(cls, tz: Optional[timezone] = None) -> datetime:
        """Return a datetime object for the current time.

        ### Arguments:
        * tz: Optional[timezone] - The timezone for the object. Defaults to None (local time).

        ### Returns:
        datetime - The current datetime object.
        """
        return datetime.now(tz=tz)

    @classmethod
    def datetime_from_timestamp(
        cls,
        timestamp_seconds: Union[int, float],
        tz: Optional[timezone] = None,
    ) -> datetime:
        """Create a datetime object from a Unix timestamp.

        ### Arguments:
        * timestamp_seconds: Union[int, float] - The Unix timestamp in seconds.
        * tz: Optional[timezone] - The timezone for the object. Defaults to None.

        ### Returns:
        datetime - The datetime object corresponding to the timestamp.

        ### Raises:
        * TypeError: If `timestamp_seconds` is not an int or float.
        """
        if not isinstance(timestamp_seconds, (int, float)):
            raise Raise.error(
                f"Expected int or float type, received: '{type(timestamp_seconds)}'.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        return datetime.fromtimestamp(timestamp_seconds, tz=tz)

    @classmethod
    def elapsed_time_from_seconds(cls, seconds: Union[int, float]) -> timedelta:
        """Convert a duration in seconds into a timedelta object.

        ### Arguments:
        * seconds: Union[int, float] - The duration in seconds.

        ### Returns:
        timedelta - The timedelta object representing the duration.

        ### Raises:
        * TypeError: If `seconds` is not an int or float.
        """
        if not isinstance(seconds, (int, float)):
            raise Raise.error(
                f"Expected int or float type, received: '{type(seconds)}'.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        return timedelta(seconds=seconds)

    @classmethod
    def elapsed_time_from_timestamp(
        cls, seconds: Union[int, float], tz: Optional[timezone] = None
    ) -> timedelta:
        """Calculate the elapsed time from a given timestamp to now.

        ### Arguments:
        * seconds: Union[int, float] - The starting Unix timestamp in seconds.
        * tz: Optional[timezone] - The timezone for the calculation. Defaults to None.

        ### Returns:
        timedelta - The timedelta object for the elapsed time, accurate to the second.

        ### Raises:
        * TypeError: If `seconds` is not an int or float.
        """
        if not isinstance(seconds, (int, float)):
            raise Raise.error(
                f"Expected int or float type, received: '{type(seconds)}'.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        out: timedelta = cls.now(tz=tz) - datetime.fromtimestamp(seconds, tz=tz)
        return timedelta(days=out.days, seconds=out.seconds)


class Timestamp(NoNewAttributes):
    """A utility class for generating Unix timestamps."""

    @classmethod
    def now(
        cls, returned_type: Union[type[int], type[float]] = int
    ) -> Union[int, float]:
        """Get the current Unix timestamp.

        ### Arguments:
        * returned_type: Union[type[int], type[float]] - The desired type, `int` (default) or `float`.

        ### Returns:
        Union[int, float] - The current Unix timestamp.

        ### Raises:
        * TypeError: If `returned_type` is not `int` or `float`.
        """
        if returned_type not in (int, float):
            raise Raise.error(
                f"Expected int or float type, received: '{returned_type}'.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        if returned_type == int:
            return int(time())
        return time()

    @classmethod
    def from_string(
        cls,
        date_string: str,
        format: str,
        returned_type: Union[type[int], type[float]] = int,
    ) -> Union[int, float]:
        """Create a Unix timestamp from a string representation of a date.

        ### Arguments:
        * date_string: str - The string containing the date and/or time.
        * format: str - The `strptime` format code used to parse the `date_string`.
        * returned_type: Union[type[int], type[float]] - The desired type, `int` (default) or `float`.

        ### Returns:
        Union[int, float] - The Unix timestamp derived from the string.

        ### Raises:
        * TypeError: If `returned_type` is not `int` or `float`.
        * ValueError: If `date_string` cannot be parsed with the given `format`.
        """
        if returned_type not in (int, float):
            raise Raise.error(
                f"Expected int or float type, received: '{returned_type}'.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )

        try:
            element: datetime = datetime.strptime(date_string, format)
        except ValueError as ex:
            raise Raise.error(f"{ex}", ValueError, cls.__qualname__, currentframe())

        if returned_type == int:
            return int(datetime.timestamp(element))
        return datetime.timestamp(element)

    @classmethod
    def month_timestamp_tuple(
        cls,
        query_date: Optional[Union[float, int, datetime]] = None,
        tz: Optional[timezone] = timezone.utc,
    ) -> Tuple[float, float]:
        """Get the start and end Unix timestamps for a given month.

        If `query_date` is not provided, the current month is used.

        ### Arguments:
        * query_date: Optional[Union[float, int, datetime]] - The date to determine the month (timestamp or datetime object). Defaults to None.
        * tz: Optional[timezone] - The timezone for the calculation. Defaults to `timezone.utc`.

        ### Returns:
        Tuple[float, float] - A tuple with the start and end timestamps of the month.

        ### Raises:
        * TypeError: If `tz` or `query_date` is an unsupported type.
        """

        q_date: datetime = DateTime.now(tz)

        # check types
        if tz and not isinstance(tz, timezone):
            raise Raise.error(
                f"Expected tzinfo as datetime.timezone type, received: '{type(tz)}'.",
                TypeError,
                cls.__qualname__,
                currentframe(),
            )
        if query_date:
            if isinstance(query_date, (int, float)):
                # receive timestamp number
                q_date = DateTime.datetime_from_timestamp(
                    timestamp_seconds=query_date, tz=tz
                )
            elif isinstance(query_date, datetime):
                # receive datetime format
                q_date = query_date
            else:
                raise Raise.error(
                    f"Excepted date as timestamp or datetime.datetime format, received: '{type(query_date)}'.",
                    TypeError,
                    cls.__qualname__,
                    currentframe(),
                )

        # Call the helper method with the determined year and month
        return cls._get_month_timestamp(q_date.year, q_date.month, tz=tz)

    @classmethod
    def _get_month_timestamp(
        cls, year: int, month: int, tz: Optional[timezone] = timezone.utc
    ) -> Tuple[float, float]:
        """Generate the start and end Unix timestamps for a specific month and year.

        This is a private helper method.

        ### Arguments:
        * year: int - The target year.
        * month: int - The target month (1-12).
        * tz: Optional[timezone] - The timezone for the calculation. Defaults to `timezone.utc`.

        ### Returns:
        Tuple[float, float] - A tuple with the start and end timestamps of the month.

        ### Raises:
        * ValueError: If the month is not between 1 and 12.
        """
        # Validate month input
        if not 1 <= month <= 12:
            raise Raise.error(
                "Month must be between 1 and 12.",
                ValueError,
                cls.__qualname__,
                currentframe(),
            )

        # The first moment of the month in UTC
        start_dt = datetime(year, month, 1, 0, 0, 0, tzinfo=tz)

        # Find the number of days in the month
        _, num_days = calendar.monthrange(year, month)

        # The very last moment of the month in UTC
        end_dt = datetime(year, month, num_days, 23, 59, 59, 999999, tzinfo=tz)

        # Convert datetime objects to Unix timestamps
        start_timestamp: float = start_dt.timestamp()
        end_timestamp: float = end_dt.timestamp()

        return (start_timestamp, end_timestamp)


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 19.09.2023

Purpose: JskToolBox - Comprehensive Python 3 library for building robust applications.

A collection of utility modules providing base classes, configuration management,
networking tools, logging infrastructure, and device-specific integrations.
"""

from typing import Tuple

__author__ = "Jacek 'Szumak' Kotlarski"
__version_info__: Tuple[int, int, int] = (1, 2, 3)  # Major, Minor, Patch
__suffix__: str = ""  # Optional suffix for pre-release versions
__version__: str = ".".join(map(str, __version_info__)) + __suffix__



# -*- coding: UTF-8 -*-
"""
Author:  Jacek Kotlarski --<szumak@virthost.pl>
Created: 08.12.2023

Purpose: Query class as helper for build query elements dict.
"""

from copy import copy
from typing import Dict, Optional

from .....basetool.data import BData
from .....attribtool import ReadOnlyClass


class RBQuery(BData):
    """RBQuery class helper."""

    class Keys(object, metaclass=ReadOnlyClass):
        """Immutable keys for RBQuery internal storage."""

        SEARCH = "_search_query_"

    def __init__(self) -> None:
        """Constructor."""
        self._set_data(
            key=RBQuery.Keys.SEARCH,
            set_default_type=Dict,
            value={},
        )

    def add_attrib(self, attrib: str, value: Optional[str] = None) -> None:
        """Build query

        ### Arguments:
        * attrib -- name of the attribute being searched for

        ### Keyword Arguments:
        * value -- optional value of the attribute being searched for
          (default: {None})
        """
        self._get_data(key=RBQuery.Keys.SEARCH)[attrib] = value  # type: ignore

    @property
    def query(self) -> Dict:
        """Returns query dict.

        ### Returns:
        Dict - Copy of the search query dictionary.
        """
        return copy(self._get_data(key=RBQuery.Keys.SEARCH))  # type: ignore


# #[EOF]#######################################################################



# -*- coding: utf-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 2024-10-08

Purpose: EDSM API client classes.
"""

import requests  # type: ignore
import json

from typing import Dict, List, Optional, Any, Union
from requests.utils import requote_uri  # type: ignore
from inspect import currentframe

from .edsm_keys import EdsmKeys


from ..basetool.data import BData
from ..attribtool import ReadOnlyClass
from ..raisetool import Raise
from ..edmctool.stars import StarsSystem


class _Keys(object, metaclass=ReadOnlyClass):
    """Internal  keys container class."""

    OPTIONS: str = "__options__"
    SYSTEMS_URL: str = "__systems_url__"
    SYSTEM_URL: str = "__system_url__"


class Url(BData):
    """Url.

    Class for serving HTTP/HTTPS requests.
    """

    def __init__(self) -> None:
        """Create Url helper object."""
        self.__options = {
            EdsmKeys.SHOW_ID: 1,
            EdsmKeys.SHOW_PERMIT: 1,
            EdsmKeys.SHOW_COORDINATES: 1,
            EdsmKeys.SHOW_INFORMATION: 0,
            EdsmKeys.SHOW_PRIMARY_STAR: 0,
            EdsmKeys.INCLUDE_HIDDEN: 0,
        }
        self._set_data(
            key=_Keys.SYSTEMS_URL,
            value="https://www.edsm.net/api-v1/",
            set_default_type=str,
        )
        self._set_data(
            key=_Keys.SYSTEM_URL,
            value="https://www.edsm.net/api-system-v1/",
            set_default_type=str,
        )

    @property
    def __options(self) -> Dict:
        """Return the options dictionary for URL parameters.

        ### Returns:
        Dict - URL query options mapping.
        """
        return self._get_data(key=_Keys.OPTIONS)  # type: ignore

    @__options.setter
    def __options(self, value: Optional[Dict]) -> None:
        """Assign the options dictionary.

        ### Arguments:
        * value: Optional[Dict] - Options dictionary or None to initialize empty.
        """
        if value is None:
            self._set_data(key=_Keys.OPTIONS, value={}, set_default_type=Dict)
        else:
            self._set_data(key=_Keys.OPTIONS, value=value, set_default_type=Dict)

    @property
    def __system_url(self) -> str:
        """Return the system API endpoint URL.

        ### Returns:
        str - URL for single system queries.
        """
        return self._get_data(key=_Keys.SYSTEM_URL)  # type: ignore

    @property
    def __systems_url(self) -> str:
        """Return the systems API endpoint URL.

        ### Returns:
        str - URL for multiple systems queries.
        """
        return self._get_data(key=_Keys.SYSTEMS_URL)  # type: ignore

    @property
    def options(self) -> str:
        """Get url options string.

        ### Returns:
        str - Formatted URL parameter string.
        """
        out: str = ""
        for key, value in self.__options.items():
            out += f"&{key}={value}"
        return out

    def bodies_url(self, s_system: StarsSystem) -> str:
        """Returns proper API url for getting bodies information data.

        ### Arguments:
        * s_system: StarsSystem - System object containing name or address.

        ### Returns:
        str - Formatted URL for bodies query.

        ### Raises:
        * TypeError: If s_system is not a StarsSystem instance.
        """
        if not isinstance(s_system, StarsSystem):
            raise Raise.error(
                f"StarsSystem type expected, '{type(s_system)}' received",
                TypeError,
                self._c_name,
                currentframe(),
            )

        if s_system.name:
            return requote_uri(f"{self.__system_url}bodies?systemName={s_system.name}")
        if s_system.address:
            return requote_uri(f"{self.__system_url}bodies?systemId={s_system.address}")
        return ""

    def system_url(self, s_system: StarsSystem) -> str:
        """Returns proper API url for getting system data.

        ### Arguments:
        * s_system: StarsSystem - System object containing name.

        ### Returns:
        str - Formatted URL for system query.

        ### Raises:
        * TypeError: If s_system is not a StarsSystem instance.
        """
        if not isinstance(s_system, StarsSystem):
            raise Raise.error(
                f"StarsSystem type expected, '{type(s_system)}' received",
                TypeError,
                self._c_name,
                currentframe(),
            )

        if s_system.name:
            return requote_uri(
                f"{self.__systems_url}system?systemName={s_system.name}{self.options}"
            )
        return ""

    def radius_url(self, s_system: StarsSystem, radius: int) -> str:
        """Returns proper API url for getting systems data in radius.

        ### Arguments:
        * s_system: StarsSystem - System object containing name.
        * radius: int - Search radius in light years (clamped to 5-100).

        ### Returns:
        str - Formatted URL for sphere-systems query.

        ### Raises:
        * TypeError: If s_system is not a StarsSystem instance.
        """
        if not isinstance(s_system, StarsSystem):
            raise Raise.error(
                f"StarsSystem type expected, '{type(s_system)}' received",
                TypeError,
                self._c_name,
                currentframe(),
            )
        if not isinstance(radius, int):
            radius = 50
        else:
            if radius < 5:
                radius = 5
            elif radius > 100:
                radius = 100

        if s_system.name:
            return requote_uri(
                f"{self.__systems_url}sphere-systems?systemName={s_system.name}&radius={radius}{self.options}"
            )
        return ""

    def cube_url(self, s_system: StarsSystem, size: int) -> str:
        """Returns proper API url for getting systems data in cube.

        ### Arguments:
        * s_system: StarsSystem - System object containing name.
        * size: int - Cube size in light years (clamped to 10-200).

        ### Returns:
        str - Formatted URL for cube-systems query.

        ### Raises:
        * TypeError: If s_system is not a StarsSystem instance.
        """
        if not isinstance(s_system, StarsSystem):
            raise Raise.error(
                f"StarsSystem type expected, '{type(s_system)}' received",
                TypeError,
                self._c_name,
                currentframe(),
            )
        if not isinstance(size, int):
            size = 100
        else:
            if size < 10:
                size = 10
            elif size > 200:
                size = 200

        if s_system.name:
            return requote_uri(
                f"{self.__systems_url}cube-systems?systemName={s_system.name}&size={size}{self.options}"
            )
        return ""

    def system_query(self, s_system: StarsSystem) -> Optional[Dict]:
        """Returns result of query for system data.

        ### Arguments:
        * s_system: StarsSystem - System object containing name.

        ### Returns:
        Optional[Dict] - System data from EDSM API or None on failure.

        ### Raises:
        * TypeError: If s_system is not a StarsSystem instance.
        """
        if not isinstance(s_system, StarsSystem):
            raise Raise.error(
                f"StarsSystem type expected, '{type(s_system)}' received",
                TypeError,
                self._c_name,
                currentframe(),
            )
        url: str = self.system_url(s_system)
        if not url:
            return None

        try:
            response: requests.Response = requests.get(url, timeout=30)
            if response.status_code != 200:
                print(f"Error calling API for system data: {response.status_code}")
                return None
            return json.loads(response.text)
        except Exception as ex:
            print(ex)
        return None

    def url_query(self, url: str) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
        """Returns result of query for url.

        ### Arguments:
        * url: str - Full API URL to query.

        ### Returns:
        Union[List[Dict[str, Any]], Dict[str, Any]] - Response data from EDSM API or empty list on failure.
        """
        out = []
        if not url:
            return out

        try:
            response: requests.Response = requests.get(url, timeout=60)
            if response.status_code != 200:
                print(f"Error calling API for EDSM data: {response.status_code}")
            else:
                out = json.loads(response.text)
        except Exception as ex:
            print(ex)
        return out


# #[EOF]#######################################################################



"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 2024-05-12

Purpose: Group RouterOS-specific device abstractions and helpers under a
dedicated namespace.

The subpackage collects base classes, element definitions, and top-level device
facades for MikroTik hardware while delegating networking concerns to
`jsktoolbox.devices.network`.
"""

# #[EOF]#######################################################################



# -*- coding: utf-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 2024-01-15

Purpose: Provide foundational mixins for the logging subsystem.

Defines base mixins for logger queues, engine metadata, and formatting
behaviour leveraged by higher-level logging utilities.
"""
from typing import Optional, List, Any

from ..logstool.keys import LogKeys
from ..logstool.queue import LoggerQueue

from .data import BData
from ..attribtool import NoDynamicAttributes


class BLoggerQueue(BData):
    """Base mixin that exposes a lazily-created logging queue."""

    @property
    def logs_queue(self) -> Optional[LoggerQueue]:
        """Return the configured logging queue instance.

        ### Returns:
        [Optional[LoggerQueue]] - Logger queue or None when not set.
        """
        return self._get_data(
            key=LogKeys.QUEUE,
            default_value=None,
        )

    @logs_queue.setter
    def logs_queue(self, obj: Optional[LoggerQueue]) -> None:
        """Assign the logger queue instance.

        ### Arguments:
        * obj: Optional[LoggerQueue] - Queue instance or None.
        """
        self._set_data(
            key=LogKeys.QUEUE, value=obj, set_default_type=Optional[LoggerQueue]
        )


class BLoggerEngine(BData):
    """Base mixin for logger engine metadata."""

    @property
    def name(self) -> Optional[str]:
        """Return the configured application name.

        ### Returns:
        [Optional[str]] - Application name value or None.
        """
        return self._get_data(
            key=LogKeys.NAME,
            default_value=None,
        )

    @name.setter
    def name(self, value: str) -> None:
        """Assign the application name.

        ### Arguments:
        * value: str - Application name string.
        """
        self._set_data(key=LogKeys.NAME, value=value, set_default_type=Optional[str])


class BLogFormatter(NoDynamicAttributes):
    """Base mixin for log formatters leveraging simple templates."""

    __blf_template: Optional[str] = None
    __blf_forms: Optional[List] = None

    def format(self, message: str, name: Optional[str] = None) -> str:
        """Render a log message based on the configured forms list.

        ### Arguments:
        * message: str - Log string to include in the output.
        * name: Optional[str] - Optional application name.

        ### Returns:
        [str] - Formatted log payload.
        """
        out: str = ""
        for item in self._forms_:
            if callable(item):
                out += f"{item()} "
            elif isinstance(item, str):
                if name is None:
                    if item.find("name") == -1:
                        out += item.format(message=f"{message}")
                else:
                    if item.find("name") > 0:
                        out += item.format(
                            name=f"{name}",
                            message=f"{message}",
                        )
        return out

    @property
    def _forms_(self) -> List:
        """Return the list of formatter components.

        ### Returns:
        [List] - Current forms definition list.
        """
        if self.__blf_forms is None:
            self.__blf_forms = []
        return self.__blf_forms

    @_forms_.setter
    def _forms_(self, item: Any) -> None:
        """Append a formatter component to the forms list.

        ### Arguments:
        * item: Any - Callable or string template.
        """
        self._forms_.append(item)


# #[EOF]#######################################################################



"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 2024-05-12

Purpose: Expose device-related subpackages (connectors, libraries, vendors)
under a unified namespace.

This module keeps the import surface minimal to avoid side effects while still
providing a single integration point for device tooling. Exports are resolved
lazily to defer heavyweight dependencies until they are explicitly requested.
"""

from importlib import import_module
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .libs.base import BDebug as BDebug
    from .libs.base import BDev as BDev
    from .libs.converters import B64Converter as B64Converter
    from .network.connectors import API as API
    from .network.connectors import IConnector as IConnector
    from .network.connectors import SSH as SSH
    from .mikrotik.base import BRouterOS as BRouterOS
    from .mikrotik.base import Element as Element
    from .mikrotik.routerboard import RouterBoard as RouterBoard

__all__ = [
    "API",
    "BDebug",
    "BDev",
    "BRouterOS",
    "B64Converter",
    "Element",
    "IConnector",
    "RouterBoard",
    # "SSH", # Not Implemented Yet
]

_EXPORT_MAP = {
    "BDebug": ("libs.base", "BDebug"),
    "BDev": ("libs.base", "BDev"),
    "B64Converter": ("libs.converters", "B64Converter"),
    "IConnector": ("network.connectors", "IConnector"),
    "API": ("network.connectors", "API"),
    # "SSH": ("network.connectors", "SSH"),
    "BRouterOS": ("mikrotik.base", "BRouterOS"),
    "Element": ("mikrotik.base", "Element"),
    "RouterBoard": ("mikrotik.routerboard", "RouterBoard"),
}


def __getattr__(name: str) -> Any:
    """Resolve configured device exports on demand.

    ### Arguments:
    * name: str - Requested attribute name.

    ### Returns:
    [Any] - Resolved attribute from the target submodule.

    ### Raises:
    * AttributeError: Raised when the attribute is not registered.
    """
    if name not in _EXPORT_MAP:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")
    module_name, attr_name = _EXPORT_MAP[name]
    module = import_module(f"{__name__}.{module_name}")
    value = getattr(module, attr_name)
    globals()[name] = value
    return value


def __dir__() -> list[str]:
    """Expose lazy exports to dir()."""
    return sorted(__all__)


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 04.12.2023

Purpose: MikroTik RouterOS main class.
"""

from typing import Dict, List, Optional, Union, Tuple, Any
from inspect import currentframe

from ...logstool.queue import LoggerQueue

from ...attribtool import NoDynamicAttributes, ReadOnlyClass
from ...raisetool import Raise
from ...logstool.logs import LoggerClient


from ...netaddresstool.ipv4 import (
    Address,
    Netmask,
    Network,
    SubNetwork,
)
from ...netaddresstool.ipv6 import (
    Address6,
    Network6,
    Prefix6,
    SubNetwork6,
)

from .base import BRouterOS, Element
from ..network.connectors import IConnector
from .elements.libs.interfaces import IElement
from .elements.certificate import RBCertificate
from .elements.disk import RBDisk
from .elements.file import RBFile
from .elements.interface import RBInterface
from .elements.ip import RBIp
from .elements.ipv6 import RBIpv6
from .elements.lcd import RBLcd
from .elements.log import RBLog
from .elements.mpls import RBMpls
from .elements.partitions import RBPartitions
from .elements.port import RBPort
from .elements.ppp import RBPpp
from .elements.queue import RBQueue
from .elements.radius import RBRadius
from .elements.routing import RBRouting
from .elements.snmp import RBSnmp
from .elements.system import RBSystem
from .elements.tool import RBTool
from .elements.user import RBUser


class _Elements(object, metaclass=ReadOnlyClass):
    """Internal keys class."""

    CERTIFICATE: str = "certificate"
    DISK: str = "disk"
    FILE: str = "file"
    INTERFACE: str = "interface"
    IP: str = "ip"
    IPV6: str = "ipv6"
    LCD: str = "lcd"
    LOG: str = "log"
    MPLS: str = "mpls"
    PARTITIONS: str = "partitions"
    PORT: str = "port"
    PPP: str = "ppp"
    QUEUE: str = "queue"
    RADIUS: str = "radius"
    ROUTING: str = "routing"
    SNMP: str = "snmp"
    SYSTEM: str = "system"
    TOOL: str = "tool"
    USER: str = "user"


class RouterBoard(BRouterOS):
    """MikroTik RouterBoard class."""

    system = None

    def __init__(
        self,
        connector: IConnector,
        qlog: Optional[LoggerQueue] = None,
        debug: bool = False,
        verbose: bool = False,
    ) -> None:
        """Constructor.

        ### Arguments:
        * connector: IConnector - Network connector interface for RouterOS communication.
        * qlog: Optional[LoggerQueue] - Logger queue for logging operations. Defaults to None.
        * debug: bool - Enable debug mode. Defaults to False.
        * verbose: bool - Enable verbose output. Defaults to False.
        """
        super().__init__(
            None,
            connector,
            LoggerClient(queue=qlog, name=self._c_name),
            debug,
            verbose,
        )
        self.root = "/"

        # add elements
        if self._ch is None:
            return None
        self.elements[_Elements.CERTIFICATE] = RBCertificate(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.DISK] = RBDisk(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.FILE] = RBFile(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.INTERFACE] = RBInterface(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.IP] = RBIp(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.IPV6] = RBIpv6(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.LCD] = RBLcd(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.LOG] = RBLog(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.MPLS] = RBMpls(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.PARTITIONS] = RBPartitions(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.PORT] = RBPort(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.PPP] = RBPpp(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.QUEUE] = RBQueue(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.RADIUS] = RBRadius(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.ROUTING] = RBRouting(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.SNMP] = RBSnmp(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.SYSTEM] = RBSystem(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.TOOL] = RBTool(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )
        self.elements[_Elements.USER] = RBUser(
            parent=self,
            connector=self._ch,
            qlog=self.logs.logs_queue if self.logs is not None else None,
            debug=self.debug,
            verbose=self.verbose,
        )


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 04.12.2023

Purpose: Base classes for RouterOS
"""

from inspect import currentframe
from typing import Dict, List, Optional, Union, TypeVar, Any

from jsktoolbox.raisetool import Raise

from ...logstool.queue import LoggerQueue

from ...attribtool import ReadOnlyClass
from ...logstool.logs import LoggerClient
from .elements.libs.base import BElement
from ..libs.base import BDev
from ..network.connectors import IConnector


class _Keys(object, metaclass=ReadOnlyClass):
    """Keys definition class.

    For internal purpose only.
    """

    ELEMENTS: str = "__elements__"
    LOADED: str = "__loaded__"


class BRouterOS(BDev, BElement):
    """Base class for RouterOS."""

    def __init__(
        self,
        parent: Optional[BDev],
        connector: IConnector,
        logs: LoggerClient,
        debug: bool,
        verbose: bool,
    ) -> None:
        """Constructor.

        ### Arguments:
        * parent: Optional[BDev] - Parent device object. Can be None for root devices.
        * connector: IConnector - Network connector interface for communication.
        * logs: LoggerClient - Logger client for logging operations.
        * debug: bool - Enable debug mode.
        * verbose: bool - Enable verbose output.
        """
        self.parent = parent
        self._ch = connector
        self.logs = logs
        self.debug = debug
        self.verbose = verbose
        self._set_data(key=_Keys.LOADED, set_default_type=bool, value=False)
        self._set_data(key=_Keys.ELEMENTS, set_default_type=Dict, value={})

    def __str__(self) -> str:
        """Returns a string representing the object."""
        return (
            f"{self._c_name}"
            f"(path='{self.root}', "
            f"elements='{self.elements}', "
            f"attrib='{self.attrib}', "
            f"list='{self.list}'"
            f")"
        )

    def _add_elements(self, parent: "BRouterOS", elements_dict: Dict) -> None:
        """Add children from configuration dict."""
        if parent._ch is None:
            return None
        if not isinstance(elements_dict, Dict):
            raise Raise.error(
                f"Expected dictionary type, received: {type(elements_dict)}",
                TypeError,
                self._c_name,
                currentframe(),
            )
        for key in elements_dict.keys():
            if key in self.elements:
                # duplicate
                if self.debug and self.logs is not None:
                    self.logs.message_debug = f'duplicate key found: "{key}"'
                    continue
            queue = parent.logs.logs_queue if parent.logs is not None else None
            obj = Element(
                key=key,
                parent=parent,
                connector=parent._ch,
                qlog=queue,
                debug=parent.debug,
                verbose=parent.verbose,
            )
            self.elements[key] = obj
            if elements_dict[key]:
                obj._add_elements(obj, elements_dict[key])

    def dump(self) -> None:
        """Dump all dataset.

        For developer debug purpose only.
        """
        print(self.root)
        self.load(self.root)
        if self.attrib:
            print(f"attrib: {self.attrib}")
        if self.list:
            for item in self.list:
                print(f"list: {item}")
        for item in self.elements.values():
            item.dump()

    def element(
        self,
        root: str,
        auto_load: bool = False,
    ) -> Optional["Element"]:
        """Returns the Element object for corresponding path."""
        # check if first and last char in path is '/'
        if root:
            if root[0] != "/":
                root = f"/{root}"
            if root[-1:] != "/":
                root = f"{root}/"
        for key in self.elements.keys():
            element: Element = self.elements[key]
            if element.root == root:
                if auto_load:
                    element.load(root)
                return element  # type: ignore
            element2: Element = element.element(  # type: ignore
                root,
                auto_load,
            )
            if element2 is not None:
                return element2  # type: ignore
        return None

    @property
    def elements(self) -> Dict[str, Any]:
        """Return elements dict.

        ### Returns:
        Dict[str, Any] - Dictionary containing element objects indexed by their names.
        """
        return self._get_data(key=_Keys.ELEMENTS)  # type: ignore

    def get(self) -> bool:
        """Gets config for current element."""
        return self.load(self.root)

    @property
    def is_loaded(self) -> bool:
        """Returns True if loaded.

        ### Returns:
        bool - True if configuration has been loaded, False otherwise.
        """
        return self._get_data(key=_Keys.LOADED)  # type: ignore

    def load(self, root: str) -> bool:
        """Gets element config from RB."""
        if self._ch is None:
            return False
        if root is not None and not self.is_loaded:
            ret: bool = self._ch.execute(f"{root}print")
            if ret:
                out, err = self._ch.outputs()
                if (
                    out[0]
                    and isinstance(out[0], List)
                    and len(out[0]) == 1
                    and isinstance(out[0][0], Dict)
                ):
                    self.attrib.update(out[0][0])
                    self._set_data(key=_Keys.LOADED, value=True)
                elif (
                    out[0]
                    and isinstance(out[0], List)
                    and len(out[0]) > 1
                    and isinstance(out[0][0], Dict)
                ):
                    for item in out[0]:
                        self.list.append(item)
                    self._set_data(key=_Keys.LOADED, value=True)
                else:
                    if out[0]:
                        print(f"DEBUG_: {out}")
                if err[0] and self.logs is not None:
                    self.logs.message_warning = f"{out[0][0]}"
                    return False
                return True
        return False


class Element(BRouterOS):
    """MikroTik Element class."""

    # TODO:
    # methods:
    # - add
    # - set
    # - remove
    # - move
    # - is_dirty
    # - commit
    # - _reload
    # with content verification,
    # ROS version validation
    # string encoding

    def __init__(
        self,
        key: str,
        parent: BDev,
        connector: IConnector,
        qlog: Optional[LoggerQueue] = None,
        debug: bool = False,
        verbose: bool = False,
    ) -> None:
        """Constructor.

        ### Arguments:
        * key: str - Unique identifier key for this element.
        * parent: BDev - Parent device object.
        * connector: IConnector - Network connector interface for communication.
        * qlog: Optional[LoggerQueue] - Logger queue for logging operations. Defaults to None.
        * debug: bool - Enable debug mode. Defaults to False.
        * verbose: bool - Enable verbose output. Defaults to False.
        """
        super().__init__(
            parent,
            connector,
            LoggerClient(queue=qlog, name=key),
            debug,
            verbose,
        )
        self.root = f"{key}/"

    def search(
        self,
        search_dict: Dict,
    ) -> Optional[Union[List[str], Dict[str, Any]]]:
        """Returns optional Dict or List[Dict] with found results."""
        # search_dict = {
        # key1: value1,
        # key2: None,
        # }
        if self.attrib:
            test = True
            for key, value in search_dict.items():
                if value:
                    if key in self.attrib and self.attrib[key] != value:
                        test = False
                elif key not in self.attrib:
                    test = False
            if test:
                return self.attrib
            return None

        if self.list:
            out = []
            for item in self.list:
                test = True
                for key, value in search_dict.items():
                    if value:
                        if key in item and item[key] != value:
                            test = False
                    elif key not in item:
                        test = False
                if test:
                    out.append(item)
            if out:
                return out

        return None


# #[EOF]#######################################################################



# -*- coding: utf-8 -*-
"""
logs2.py
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 16.10.2025, 20:55:52

Purpose: Logging server and component mixin for EDMC toolbox.
Description:
  This module provides a logging server that manages logging destinations
  and a mixin for components to easily attach to the logging server.
"""

from __future__ import annotations
from inspect import currentframe

from re import DEBUG, L
import threading
import time
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, List, Optional, cast, Union
from venv import logger

from ..attribtool import ReadOnlyClass
from ..basetool.data import BData
from ..logstool import (
    LoggerClient,
    LoggerEngine,
    LoggerEngineFile,
    LoggerEngineStdout,
    LogFormatterDateTime,
    LogsLevelKeys,
    ThLoggerProcessor,
)
from ..basetool import BLoggerQueue
from ..raisetool import Raise


if TYPE_CHECKING:
    from ..logstool import LoggerQueue


class _Keys(object, metaclass=ReadOnlyClass):
    """Predefined keys for logging configuration."""

    APP_NAME: str = "__app_name__"
    LEC: str = "__logger_engine_config__"
    LOGGER_CLIENT: str = "__logger_client__"
    LOG_PATH: str = "__log_path__"


class LoggerEngineConfig(BData):
    """Configuration data for logger engine destinations."""

    class _LECKeys(object, metaclass=ReadOnlyClass):
        """Predefined keys for logger engine configuration."""

        BACKUP_COUNT: str = "backup_count"
        CONSOLE_LEVELS: str = "console_levels"
        CONSOLE_OBJ: str = "__console_obj__"
        FILE_LEVELS: str = "file_levels"
        FILE_OBJ: str = "__file_obj__"
        MAX_FILE_SIZE: str = "max_file_size"
        DEBUG: str = "debug_mode"

    class Console(BData):
        """Console logging configuration."""

        def __init__(self) -> None:
            """Constructor."""
            self._set_data(
                key=LoggerEngineConfig._LECKeys.CONSOLE_LEVELS,
                set_default_type=str,
                value=LogsLevelKeys.INFO,
            )

        @property
        def levels(self) -> str:
            """Return the configured console log levels.

            ### Returns:
            str - log levels as str.
            """
            level = self._get_data(
                key=LoggerEngineConfig._LECKeys.CONSOLE_LEVELS,
            )
            return level or LogsLevelKeys.INFO

        @levels.setter
        def levels(self, obj: str) -> None:
            """Assign the console log levels.

            ### Arguments:
            * obj: str - List of log levels or None.
            """
            self._set_data(
                key=LoggerEngineConfig._LECKeys.CONSOLE_LEVELS,
                value=obj,
                set_default_type=str,
            )

    class File(BData):
        """File logging configuration."""

        def __init__(self) -> None:
            """Constructor."""
            self._set_data(
                key=LoggerEngineConfig._LECKeys.FILE_LEVELS,
                set_default_type=str,
                value=LogsLevelKeys.INFO,
            )
            self._set_data(
                key=LoggerEngineConfig._LECKeys.MAX_FILE_SIZE,
                set_default_type=Optional[int],
                value=None,
            )
            self._set_data(
                key=LoggerEngineConfig._LECKeys.BACKUP_COUNT,
                set_default_type=int,
                value=0,
            )

        @property
        def levels(self) -> str:
            """Return the configured file log levels.

            ### Returns:
            str - Log levels as str.
            """
            level = self._get_data(
                key=LoggerEngineConfig._LECKeys.FILE_LEVELS,
            )
            return level or LogsLevelKeys.INFO

        @levels.setter
        def levels(self, obj: str) -> None:
            """Assign the file log levels.

            ### Arguments:
            * obj: str - log levels as str.
            """
            self._set_data(
                key=LoggerEngineConfig._LECKeys.FILE_LEVELS,
                value=obj,
                set_default_type=str,
            )

        @property
        def max_file_size(self) -> Optional[int]:
            """Return the maximum file size in bytes.

            ### Returns:
            Optional[int] - Maximum file size in bytes or None.
            """
            return self._get_data(
                key=LoggerEngineConfig._LECKeys.MAX_FILE_SIZE,
            )

        @max_file_size.setter
        def max_file_size(self, obj: Optional[int]) -> None:
            """Assign the maximum file size in bytes.

            ### Arguments:
            * obj: Optional[int] - Maximum file size in bytes or None.
            """
            self._set_data(
                key=LoggerEngineConfig._LECKeys.MAX_FILE_SIZE,
                value=obj,
            )

        @property
        def backup_count(self) -> int:
            """Return the number of backup files to keep.

            ### Returns:
            int - Number of backup files.
            """
            count = self._get_data(
                key=LoggerEngineConfig._LECKeys.BACKUP_COUNT,
            )
            return count or 0

        @backup_count.setter
        def backup_count(self, obj: int) -> None:
            """Assign the number of backup files to keep.

            ### Arguments:
            * obj: int - Number of backup files.
            """
            self._set_data(key=LoggerEngineConfig._LECKeys.BACKUP_COUNT, value=obj)

    def __init__(self) -> None:
        """Constructor."""
        self._set_data(
            key=LoggerEngineConfig._LECKeys.CONSOLE_OBJ,
            value=LoggerEngineConfig.Console(),
            set_default_type=LoggerEngineConfig.Console,
        )
        self._set_data(
            key=LoggerEngineConfig._LECKeys.FILE_OBJ,
            value=LoggerEngineConfig.File(),
            set_default_type=LoggerEngineConfig.File,
        )

    @property
    def debug(self) -> bool:
        """Return the debug mode.

        ### Returns:
        bool - Debug mode.
        """
        debug = self._get_data(
            key=LoggerEngineConfig._LECKeys.DEBUG,
        )
        if debug is None:
            return False
        return cast(bool, debug)

    @debug.setter
    def debug(self, obj: bool) -> None:
        """Assign the debug mode.

        ### Arguments:
        * obj: bool - Debug mode.
        """
        self._set_data(
            key=LoggerEngineConfig._LECKeys.DEBUG,
            value=obj,
            set_default_type=bool,
        )

    @property
    def console(self) -> LoggerEngineConfig.Console:
        """Return the console logging configuration.

        ### Returns:
        LoggerEngineConfig.Console - Console configuration instance.
        """
        console = self._get_data(
            key=LoggerEngineConfig._LECKeys.CONSOLE_OBJ,
        )
        if console is None:
            raise Raise.error(
                "LoggerEngineConfig console not initialized.",
                RuntimeError,
                self._c_name,
                currentframe(),
            )
        return cast(LoggerEngineConfig.Console, console)

    @property
    def file(self) -> LoggerEngineConfig.File:
        """Return the file logging configuration.

        ### Returns:
        LoggerEngineConfig.File - File configuration instance.
        """
        file = self._get_data(
            key=LoggerEngineConfig._LECKeys.FILE_OBJ,
        )
        if file is None:
            raise Raise.error(
                "LoggerEngineConfig file not initialized.",
                RuntimeError,
                self._c_name,
                currentframe(),
            )
        return cast(LoggerEngineConfig.File, file)


class LoggingServer(ThLoggerProcessor):
    """Run the logging processor thread with predefined engine destinations.

    The server configures a `LoggerEngine`, registers file/stdout sinks, and exposes
    helpers to create additional `LoggerClient` instances that reuse the shared queue.
    """

    def __init__(
        self,
        application_name: str,
        log_path: Path,
        engine_config: Optional[LoggerEngineConfig] = None,
        debug: bool = False,
    ) -> None:
        """Initialise the logging server with configured destinations.

        ### Arguments:
        * application_name: str - Friendly application name used in log prefixes.
        * log_path: Path - Target file for persistent log storage.
        * engine_config: Optional[LoggerEngineConfig] - Optional configuration for the logger engine.
        * debug: bool - Enables verbose lifecycle logging when True.

        ### Returns:
        None - Constructor.
        """
        super().__init__(debug=debug)
        # self._debug = debug
        self._set_data(key=_Keys.APP_NAME, value=application_name, set_default_type=str)
        self._set_data(key=_Keys.LOG_PATH, value=log_path, set_default_type=Path)
        if engine_config is None:
            engine_config = LoggerEngineConfig()
        self._set_data(
            key=_Keys.LEC,
            value=engine_config,
            set_default_type=LoggerEngineConfig,
        )
        engine = LoggerEngine()
        self.logger_engine = engine
        self._configure_engine()
        self.logger_client = LoggerClient(engine.logs_queue, name=self._c_name)

    def _configure_engine(self) -> None:
        """Configure file and console engines attached to the logger engine."""
        engine: Optional[LoggerEngine] = self.logger_engine
        log_path: Optional[Path] = self._get_data(_Keys.LOG_PATH)
        if engine is None or log_path is None:
            raise Raise.error(
                "LoggingServer initialisation incomplete.",
                RuntimeError,
                self._c_name,
                currentframe(),
            )
        config: Optional[LoggerEngineConfig] = self._get_data(_Keys.LEC)
        if config is None:
            raise Raise.error(
                "LoggingServer engine configuration missing.",
                RuntimeError,
                self._c_name,
                currentframe(),
            )

        app: Optional[str] = self._get_data(_Keys.APP_NAME)
        if app is None:
            app = ""

        formatter = LogFormatterDateTime()
        file_engine = LoggerEngineFile(name=app, formatter=formatter, buffered=False)
        file_engine.logdir = str(log_path.parent)
        file_engine.logfile = log_path.name
        file_engine.rotation_max_bytes = config.file.max_file_size
        file_engine.rotation_backup_count = config.file.backup_count

        console_engine = LoggerEngineStdout(
            name=app, formatter=formatter, buffered=False
        )

        lvl_con = False
        lvl_file = False
        for level in (
            LogsLevelKeys.DEBUG,
            LogsLevelKeys.INFO,
            LogsLevelKeys.NOTICE,
            LogsLevelKeys.WARNING,
            LogsLevelKeys.ERROR,
            LogsLevelKeys.CRITICAL,
        ):
            if config.debug or level == config.console.levels:
                lvl_con = True
            if config.debug or level == config.file.levels:
                lvl_file = True
            if lvl_con:
                engine.add_engine(level, console_engine)
            if lvl_file:
                engine.add_engine(level, file_engine)

    @property
    def logs_queue(self) -> Optional[LoggerQueue]:
        """Expose the server queue for external clients.

        ### Returns:
        Optional[LoggerQueue] - Shared logging queue instance or None.
        """
        engine: Optional[LoggerEngine] = self.logger_engine
        if engine is None:
            return None
        return engine.logs_queue

    def create_client(self, name: Optional[str] = None) -> LoggerClient:
        """Create a logger client attached to the server queue.

        ### Arguments:
        * name: Optional[str] - Friendly component name used in log prefixes.

        ### Returns:
        LoggerClient - Configured client instance.

        ### Raises:
        * RuntimeError: When the server queue is not available.
        """
        queue: Optional[LoggerQueue] = self.logs_queue
        if queue is None:
            raise Raise.error(
                "LoggingServer not initialised.",
                RuntimeError,
                self._c_name,
                currentframe(),
            )

        return LoggerClient(queue, name=name)

    def shutdown(self, timeout: float = 5.0) -> None:
        """Stop the processor thread and wait for graceful completion.

        ### Arguments:
        * timeout: float - Maximum seconds to wait for termination.

        ### Returns:
        None - The thread is joined.
        """
        self.stop()
        self.join(timeout=timeout)


class LoggingComponentMixin(BLoggerQueue):
    """Reusable mixin that binds components to the logging server."""

    def attach_logger(self, server: LoggingServer, name: Optional[str] = None) -> None:
        """Bind the component to the server queue and create a dedicated client.

        ### Arguments:
        * server: LoggingServer - Active logging server instance.
        * name: Optional[str] - Optional override for the client name.

        ### Returns:
        None - Internal logger client configured.
        """
        queue: Optional[LoggerQueue] = server.logs_queue
        if queue is None:
            raise Raise.error(
                "Logging server queue not available.",
                RuntimeError,
                self._c_name,
                currentframe(),
            )

        self.logs_queue = queue
        resolved_name: str = name or self._c_name

        self._set_data(
            key=_Keys.LOGGER_CLIENT,
            value=LoggerClient(queue, name=resolved_name),
            set_default_type=LoggerClient,
        )

    @property
    def logger(self) -> LoggerClient:
        """Return the attached logger client.

        ### Returns:
        LoggerClient - Component-specific logger client.

        ### Raises:
        * RuntimeError: When the logger has not been attached.
        """
        logger: Optional[LoggerClient] = self._get_data(_Keys.LOGGER_CLIENT)
        if logger is None:
            raise Raise.error(
                "Logger client not attached.",
                RuntimeError,
                self._c_name,
                currentframe(),
            )

        return cast(LoggerClient, logger)


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 14.09.2023

Purpose: Classes for IPv6

https://www.ibm.com/docs/en/ts3500-tape-library?topic=formats-subnet-masks-ipv4-prefixes-ipv6
"""

import socket

from copy import deepcopy
from inspect import currentframe
import warnings
from typing import Iterator, Optional, TypeVar, Union, List

from ..attribtool import NoDynamicAttributes
from ..raisetool import Raise
from .libs.words import Word16
from ..libs.interfaces.comparators import IComparators
from ..basetool.classes import BClasses

TAddress6 = TypeVar("TAddress6", bound="Address6")
DEFAULT_IPV6_HOST_LIMIT: int = 65536
DEFAULT_IPV6_SUBNET_LIMIT: int = 4096


class Address6(IComparators, BClasses, NoDynamicAttributes):
    """Address6 class for representing IPv6 addresses.

    Constructor arguments:
    addr: Union[str, int, List[Word16]] -- IPv6 address representation as string, integer or list of eight Word16

    Public property:
    words: List[Word16] -- Return list of eight Word16

    Public setter:
    words: Union[str, int, List] -- Set IPv6 address from string, integer or list of Word16.
    """

    __var_int: int = 0

    def __init__(
        self, addr: Union[str, int, Union[List[int], List[str], List[Word16]]]
    ) -> None:
        """Constructor.

        ### Arguments:
        * addr: Union[str, int, Union[List[int], List[str], List[Word16]]] - IPv6 address in various formats (string, integer, or list of 16-bit words).
        """
        self.words = addr

    def __eq__(self, arg: Union[TAddress6, object]) -> bool:
        """Equal."""
        if not isinstance(arg, Address6):
            raise Raise.error(
                f"Expected argument of Address6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) == int(arg)

    def __ge__(self, arg: Union[TAddress6, object]) -> bool:
        """Greater or equal."""
        if not isinstance(arg, Address6):
            raise Raise.error(
                f"Expected argument of Address6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) >= int(arg)

    def __gt__(self, arg: Union[TAddress6, object]) -> bool:
        """Greater."""
        if not isinstance(arg, Address6):
            raise Raise.error(
                f"Expected argument of Address6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) > int(arg)

    def __le__(self, arg: Union[TAddress6, object]) -> bool:
        """Less or equal."""
        if not isinstance(arg, Address6):
            raise Raise.error(
                f"Expected argument of Address6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) <= int(arg)

    def __lt__(self, arg: Union[TAddress6, object]) -> bool:
        """Less."""
        if not isinstance(arg, Address6):
            raise Raise.error(
                f"Expected argument of Address6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) < int(arg)

    def __ne__(self, arg: Union[TAddress6, object]) -> bool:
        """Negative."""
        if not isinstance(arg, Address6):
            raise Raise.error(
                f"Expected argument of Address6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) != int(arg)

    @staticmethod
    def __check_groups(group_list: List[str]) -> List[str]:
        for i in range(0, len(group_list)):
            group_list[i] = group_list[i].zfill(4)
        return group_list

    @staticmethod
    def __expand_ipv6(ipv6_address: str) -> str:
        # Sprawdzenie czy adres jest już w pełnej reprezentacji
        if "::" not in ipv6_address:
            return ipv6_address

        # Podziel adres na dwie części, przed '::' i po '::'
        parts: list[str] = ipv6_address.split("::", 1)
        head: List[str] = Address6.__check_groups(parts[0].split(":"))
        tail: List[str] = Address6.__check_groups(parts[1].split(":"))

        # Oblicz ile zer należy dodać, aby osiągnąć 8 części
        missing_zeros: int = 8 - (len(head) + len(tail))
        zero_part = "0000"

        # Połącz części przed '::' i po '::', wstawiając brakujące zera
        expanded_parts: str = ":".join(head + [zero_part] * missing_zeros + tail)

        return expanded_parts

    @staticmethod
    def __is_valid_ipv6(ipv6_addr: str) -> bool:
        """Check if ipv6_addr is valid."""
        try:
            # Używamy socket.inet_pton, aby sprawdzić poprawność adresu IPv6
            socket.inet_pton(socket.AF_INET6, ipv6_addr)
            return True
        except (socket.error, ValueError):
            return False

    @staticmethod
    def __int_to_ip(ipint: int) -> str:
        """Convert ip int representation to ipv6 str."""
        # W przypadku adresu IPv6 zawsze przekształcamy go na 16 bajtów (128 bitów)
        binary_ip: bytes = ipint.to_bytes(16, byteorder="big")
        ipv6_address: str = socket.inet_ntop(socket.AF_INET6, binary_ip)

        return ipv6_address

    @staticmethod
    def __ip_to_int(ipstr: str) -> int:
        """Convert ipv6 str representation to ip int."""
        # Używamy socket.inet_pton, aby przekształcić adres IPv6 w postać binarną
        packed_ip: bytes = socket.inet_pton(socket.AF_INET6, ipstr)
        # Następnie przekształcamy binarny adres IPv6 na liczbę całkowitą (integer)
        int_ip: int = int.from_bytes(packed_ip, byteorder="big")

        return int_ip

    def __set_words_from_list(
        self, value: Union[List[int], List[str], List[Word16]]
    ) -> None:
        """Set address from list."""
        if len(value) != 8:
            raise Raise.error(
                f"Expected list of eight elements.",
                ValueError,
                self._c_name,
                currentframe(),
            )
        tmp: str = (
            f"{str(Word16(value[0]))}:"
            f"{str(Word16(value[1]))}:"
            f"{str(Word16(value[2]))}:"
            f"{str(Word16(value[3]))}:"
            f"{str(Word16(value[4]))}:"
            f"{str(Word16(value[5]))}:"
            f"{str(Word16(value[6]))}:"
            f"{str(Word16(value[7]))}"
        )
        self.__set_words_from_str(tmp)

    def __set_words_from_int(self, value: int) -> None:
        # if value >= 0 and value <= 340282366920938463463374607431768211455:
        if value in range(0, 340282366920938463463374607431768211456):
            self.__var_int = value
        else:
            raise Raise.error(
                f"IP-int out of range (0-340282366920938463463374607431768211455), received: {value}",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __set_words_from_str(self, value: str) -> None:
        if Address6.__is_valid_ipv6(value):
            self.__var_int = Address6.__ip_to_int(value)
        else:
            raise Raise.error(
                f"IPv6 address is invalid: {value}",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __int__(self) -> int:
        """Return ipv6 representation as integer."""
        return self.__var_int

    def __str__(self) -> str:
        """Return string representation of address."""
        return Address6.__int_to_ip(self.__var_int)

    def __repr__(self) -> str:
        """Return representation of object."""
        return f"{self._c_name}('{str(self)}')"

    @property
    def words(self) -> List[Word16]:
        """Return words list of eight Word16.

        ### Returns:
        List[Word16] - List of eight Word16 objects representing each 16-bit segment.
        """
        tmp: list[str] = Address6.__expand_ipv6(str(self)).split(":")
        return [
            Word16(tmp[0]),
            Word16(tmp[1]),
            Word16(tmp[2]),
            Word16(tmp[3]),
            Word16(tmp[4]),
            Word16(tmp[5]),
            Word16(tmp[6]),
            Word16(tmp[7]),
        ]

    @words.setter
    def words(
        self, value: Union[str, int, Union[List[int], List[str], List[Word16]]]
    ) -> None:
        """Set IPv6 address from string, integer or list of Word16.

        ### Arguments:
        * value: Union[str, int, Union[List[int], List[str], List[Word16]]] - IPv6 address representation.
        """
        if isinstance(value, List):
            self.__set_words_from_list(value)
        elif isinstance(value, int):
            self.__set_words_from_int(value)
        elif isinstance(value, str):
            self.__set_words_from_str(value)
        else:
            raise Raise.error(
                f"Expected String or Integer or List type, received: {type(value)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )


# prefix
# https://www.heficed.com/subnet-mask-cheat-sheet/

TPrefix6 = TypeVar("TPrefix6", bound="Prefix6")


class Prefix6(IComparators, BClasses, NoDynamicAttributes):
    """Prefix6 class for IPv6 addresses.

    Constructor argument:
    prefix: Union[str, int] -- Set prefix from string or integer.

    Public property:
    prefix: str -- Return prefix as string.

    Public setter:
    prefix: Union[str, int] -- Set prefix from string or integer.
    """

    __prefix_int: int = 0

    def __init__(self, prefix: Union[str, int]) -> None:
        """Constructor.

        ### Arguments:
        * prefix: Union[str, int] - IPv6 prefix length (0-128) as string or integer.
        """
        self.prefix = prefix

    def __eq__(self, arg: Union[TPrefix6, object]) -> bool:
        """Equal."""
        if not isinstance(arg, Prefix6):
            raise Raise.error(
                f"Expected argument of Prefix6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) == int(arg)

    def __ge__(self, arg: Union[TPrefix6, object]) -> bool:
        """Greater or equal."""
        if not isinstance(arg, Prefix6):
            raise Raise.error(
                f"Expected argument of Prefix6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) >= int(arg)

    def __gt__(self, arg: Union[TPrefix6, object]) -> bool:
        """Greater."""
        if not isinstance(arg, Prefix6):
            raise Raise.error(
                f"Expected argument of Prefix6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) > int(arg)

    def __le__(self, arg: Union[TPrefix6, object]) -> bool:
        """Less or equal."""
        if not isinstance(arg, Prefix6):
            raise Raise.error(
                f"Expected argument of Prefix6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) <= int(arg)

    def __lt__(self, arg: Union[TPrefix6, object]) -> bool:
        """Less."""
        if not isinstance(arg, Prefix6):
            raise Raise.error(
                f"Expected argument of Prefix6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) < int(arg)

    def __ne__(self, arg: Union[TPrefix6, object]) -> bool:
        """Negative."""
        if not isinstance(arg, Prefix6):
            raise Raise.error(
                f"Expected argument of Prefix6 type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) != int(arg)

    def __str__(self) -> str:
        """Return prefix as string.

        ### Returns:
        str - The prefix value as string.
        """
        return str(self.__prefix_int)

    def __int__(self) -> int:
        """Return prefix as integer."""
        return self.__prefix_int

    def __repr__(self) -> str:
        """Return Prefix6 representation string."""
        return f"{self._c_name}({int(self)})"

    def __range_validator(self, value: int) -> bool:
        """Proper range validator."""
        if value not in range(0, 129):
            raise Raise.error(
                f"Prefix out of range (0-128), received: {value}",
                ValueError,
                self._c_name,
                currentframe(),
            )
        return True

    @staticmethod
    def __is_integer(value: str) -> bool:
        try:
            int(value)
            return True
        except:
            return False

    @property
    def prefix(self) -> str:
        """Return prefix as string.

        ### Returns:
        str - The prefix value as string.
        """
        return str(self)

    @prefix.setter
    def prefix(self, value: Union[str, int]) -> None:
        """Set prefix from string or integer.

        ### Arguments:
        * value: Union[str, int] - Prefix value as integer (0-128) or string representation.
        """
        if isinstance(value, int) and self.__range_validator(value):
            self.__prefix_int = value
        elif isinstance(value, str):
            if Prefix6.__is_integer(value) and self.__range_validator(int(value)):
                self.__prefix_int = int(value)
            else:
                raise Raise.error(
                    f"Expected proper integer string, '{value}' received.",
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
        else:
            raise Raise.error(
                f"Expected String or Integer type, received: '{type(value)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )


# Network
class Network6(BClasses, NoDynamicAttributes):
    """Network6 IPv6 class.

    Constructor argument:
    addr: Union[str, List] -- Set IPv6 network address from string or two element list of address [Address6,str,int,list] and prefix [Prefix6, str, int].

    Public property:
    address: Address6 -- Return IPv6 address set in the constructor.
    count: int -- Return count hosts addresses in network range.
    hosts(limit: Optional[int] = DEFAULT_IPV6_HOST_LIMIT): List[Address6] -- Deprecated helper returning host list with an optional safety limit.
    iter_hosts(limit: Optional[int] = DEFAULT_IPV6_HOST_LIMIT) -> Iterator[Address6] -- Lazy host iterator respecting optional limits.
    network: Address6 -- Return network address.
    prefix: Prefix6 -- Return prefix.
    max: Address6 -- Return max address of host in network range.
    min: Address6 -- Return min address of host in network range.
    """

    __address: Address6 = None  # type: ignore
    __prefix: Prefix6 = None  # type: ignore

    def __init__(self, addr: Union[str, List]) -> None:
        """Constructor.

        ### Arguments:
        * addr: Union[str, List] - IPv6 network address in CIDR notation (string) or list format.
        """
        if isinstance(addr, str):
            self.__network_from_str(addr)
        elif isinstance(addr, List):
            self.__network_from_list(addr)
        else:
            raise Raise.error(
                f"Expected IP network string or list type, received: '{type(addr)}'.",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __str__(self) -> str:
        """Return string representation of network address."""
        return f"{self.network}/{int(self.prefix)}"

    def __repr__(self) -> str:
        """Return  string representation of class object."""
        return f"{self._c_name}('{str(self)}')"

    def __network_from_str(self, addr: str) -> None:
        """Build configuration from string."""
        if addr.find("/") > 0:
            tmp: list[str] = addr.split("/")
            self.__address = Address6(tmp[0])
            self.__prefix = Prefix6(tmp[1])
        else:
            raise Raise.error(
                f"Expected network address in 'ipv6/prefix' format string, received: '{addr}'.",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __network_from_list(self, addr: List) -> None:
        """Build configuration from list."""
        if len(addr) != 2:
            raise Raise.error(
                "A list of two elements was expected: ['ipv6','prefix']",
                ValueError,
                self._c_name,
                currentframe(),
            )
        if isinstance(addr[0], Address6):
            self.__address = deepcopy(addr[0])
        else:
            self.__address = Address6(addr[0])
        if isinstance(addr[1], Prefix6):
            self.__prefix = deepcopy(addr[1])
        else:
            self.__prefix = Prefix6(addr[1])

    @property
    def address(self) -> Address6:
        """Return IPv6 address.

        ### Returns:
        Address6 - The IPv6 address object.
        """
        return self.__address

    @property
    def count(self) -> int:
        """Return number of hosts in subnet.

        ### Returns:
        int - Number of addresses in the subnet.
        """
        return 2 ** (128 - int(self.prefix))

    def hosts(self, limit: Optional[int] = DEFAULT_IPV6_HOST_LIMIT) -> List[Address6]:
        """Return list of hosts addresses. (Deprecated)

        ### Arguments:
        * limit: Optional[int] - Maximum number of hosts allowed to materialise. Defaults to DEFAULT_IPV6_HOST_LIMIT.

        ### Returns:
        [List[Address6]] - Host addresses contained within the subnet.

        ### Raises:
        * ValueError: Propagated from `iter_hosts()` when the host count exceeds the configured limit.
        """
        warnings.warn(
            "Network6.hosts() is deprecated; use Network6.iter_hosts() for lazy iteration.",
            DeprecationWarning,
            stacklevel=2,
        )
        return list(self.iter_hosts(limit=limit))

    def iter_hosts(
        self, limit: Optional[int] = DEFAULT_IPV6_HOST_LIMIT
    ) -> Iterator[Address6]:
        """Yield IPv6 host addresses lazily.

        ### Arguments:
        * limit: Optional[int] - Maximum number of hosts allowed before raising. Defaults to DEFAULT_IPV6_HOST_LIMIT.

        ### Returns:
        Iterator[Address6] - Generator producing host addresses in ascending order.

        ### Raises:
        * ValueError: Raised when the host count exceeds the configured limit.
        """
        effective_limit = limit
        host_count = self.count
        if effective_limit is not None and host_count > effective_limit:
            raise Raise.error(
                (
                    f"Network6 host count ({host_count}) exceeds limit "
                    f"({effective_limit}). Use iter_hosts(limit=None) for explicit override."
                ),
                ValueError,
                self._c_name,
                currentframe(),
            )
        start = int(self.min)
        for i in range(0, host_count):
            yield Address6(start + i)

    @property
    def max(self) -> Address6:
        """Return last IPv6 address from subnet.

        ### Returns:
        Address6 - The last address in the subnet.
        """
        return Address6(int(self.network) + self.count - 1)

    @property
    def min(self) -> Address6:
        """Return first IPv6 address from subnet.

        ### Returns:
        Address6 - The first address in the subnet.
        """
        ip = int(self.address)
        mask: int = (1 << 128 - int(self.prefix)) - 1
        net: int = ip & ~mask
        subnet_address = []
        for _ in range(8):
            subnet_address.insert(0, format(net & 0xFFFF, "x"))
            net >>= 16
        return Address6(":".join(subnet_address))

    @property
    def network(self) -> Address6:
        """Return network address.

        ### Returns:
        Address6 - The network address.
        """
        return self.min

    @property
    def prefix(self) -> Prefix6:
        """Return IPv6 network prefix.

        ### Returns:
        Prefix6 - The network prefix object.
        """
        return self.__prefix


# SubNetwork
class SubNetwork6(BClasses, NoDynamicAttributes):
    """SubNetwork6 calculator class.

    Constructor argument:
    network: Network6 -- The address of the network where the subnet is being searched for.
    prefix: Prefix6 -- Subnet prefix.

    Public property:
    subnets(limit: Optional[int] = DEFAULT_IPV6_SUBNET_LIMIT): List[Network6] -- Deprecated helper returning subnet list with an optional safety limit.
    iter_subnets(limit: Optional[int] = DEFAULT_IPV6_SUBNET_LIMIT) -> Iterator[Network6] -- Lazy subnet iterator.
    """

    __network: Network6 = None  # type: ignore
    __prefix: Prefix6 = None  # type: ignore

    def __init__(self, network: Network6, prefix: Prefix6) -> None:
        """Constructor.

        ### Arguments:
        * network: Network6 - IPv6 network address object.
        * prefix: Prefix6 - Prefix object defining the subnet size.
        """
        if isinstance(network, Network6) and isinstance(prefix, Prefix6):
            if int(network.prefix) <= int(prefix):
                self.__network = network
                self.__prefix = prefix
            else:
                raise Raise.error(
                    (
                        "The network prefix must be greater then or equal to the subnet prefix you are looking for."
                        f"Received: {int(network.prefix)} and {int(prefix)}"
                    ),
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
        else:
            raise Raise.error(
                f"Argument of (Network6, Prefix6) expected, ({type(network)},{type(prefix)}) received.",
                TypeError,
                self._c_name,
                currentframe(),
            )

    def subnets(
        self, limit: Optional[int] = DEFAULT_IPV6_SUBNET_LIMIT
    ) -> List[Network6]:
        """Return subnets list. (Deprecated)

        ### Arguments:
        * limit: Optional[int] - Maximum number of subnetworks allowed before raising. Defaults to DEFAULT_IPV6_SUBNET_LIMIT.

        ### Returns:
        [List[Network6]] - Generated IPv6 subnetworks.

        ### Raises:
        * ValueError: Propagated from `iter_subnets()` when the subnet count exceeds the configured limit.
        """
        warnings.warn(
            "SubNetwork6.subnets() is deprecated; use SubNetwork6.iter_subnets() for lazy iteration.",
            DeprecationWarning,
            stacklevel=2,
        )
        return list(self.iter_subnets(limit=limit))

    def iter_subnets(
        self, limit: Optional[int] = DEFAULT_IPV6_SUBNET_LIMIT
    ) -> Iterator[Network6]:
        """Yield IPv6 subnetworks lazily.

        ### Arguments:
        * limit: Optional[int] - Maximum number of subnetworks allowed before raising. Defaults to DEFAULT_IPV6_SUBNET_LIMIT.

        ### Returns:
        Iterator[Network6] - Generator producing IPv6 subnetworks in ascending order.

        ### Raises:
        * ValueError: Raised when the subnet count exceeds the configured limit.
        """
        effective_limit = limit
        produced = 0
        net_start = int(self.__network.min)
        net_end = int(self.__network.max)
        start: int = net_start
        while True:
            if effective_limit is not None and produced >= effective_limit:
                raise Raise.error(
                    (
                        f"Subnet count exceeds limit ({effective_limit}). "
                        "Use iter_subnets(limit=None) for explicit override."
                    ),
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
            subnet = Network6([Address6(start), self.__prefix])
            yield subnet
            produced += 1
            if int(subnet.max) >= net_end:
                break
            start = int(subnet.max) + 1


# #[EOF]#######################################################################



"""
Author:  Jacek Kotlarski --<szumak@virthost.pl>
Created: 2023-09-04

Purpose: Aggregate logging toolkit components (engines, formatters, queues).

Exports are loaded lazily to avoid circular import with `jsktoolbox.basetool`.
Type hints are provided for IDE support while maintaining lazy loading behavior.
"""

from importlib import import_module
from typing import TYPE_CHECKING, Any

# Type stubs for IDE support - these imports are only evaluated by type checkers
if TYPE_CHECKING:
    from .keys import LogKeys as LogKeys
    from .keys import LogsLevelKeys as LogsLevelKeys
    from .keys import SysLogKeys as SysLogKeys
    from .queue import LoggerQueue as LoggerQueue
    from .formatters import LogFormatterNull as LogFormatterNull
    from .formatters import LogFormatterDateTime as LogFormatterDateTime
    from .formatters import LogFormatterTime as LogFormatterTime
    from .formatters import LogFormatterTimestamp as LogFormatterTimestamp
    from .engines import LoggerEngineStdout as LoggerEngineStdout
    from .engines import LoggerEngineStderr as LoggerEngineStderr
    from .engines import LoggerEngineFile as LoggerEngineFile
    from .engines import LoggerEngineSyslog as LoggerEngineSyslog
    from .logs import LoggerClient as LoggerClient
    from .logs import LoggerEngine as LoggerEngine
    from .logs import ThLoggerProcessor as ThLoggerProcessor

__all__ = [
    "LogKeys",
    "LogsLevelKeys",
    "SysLogKeys",
    "LoggerQueue",
    "LogFormatterNull",
    "LogFormatterDateTime",
    "LogFormatterTime",
    "LogFormatterTimestamp",
    "LoggerEngineStdout",
    "LoggerEngineStderr",
    "LoggerEngineFile",
    "LoggerEngineSyslog",
    "LoggerClient",
    "LoggerEngine",
    "ThLoggerProcessor",
]

_EXPORT_MAP = {
    "LogKeys": ("keys", "LogKeys"),
    "LogsLevelKeys": ("keys", "LogsLevelKeys"),
    "SysLogKeys": ("keys", "SysLogKeys"),
    "LoggerQueue": ("queue", "LoggerQueue"),
    "LogFormatterNull": ("formatters", "LogFormatterNull"),
    "LogFormatterDateTime": ("formatters", "LogFormatterDateTime"),
    "LogFormatterTime": ("formatters", "LogFormatterTime"),
    "LogFormatterTimestamp": ("formatters", "LogFormatterTimestamp"),
    "LoggerEngineStdout": ("engines", "LoggerEngineStdout"),
    "LoggerEngineStderr": ("engines", "LoggerEngineStderr"),
    "LoggerEngineFile": ("engines", "LoggerEngineFile"),
    "LoggerEngineSyslog": ("engines", "LoggerEngineSyslog"),
    "LoggerClient": ("logs", "LoggerClient"),
    "LoggerEngine": ("logs", "LoggerEngine"),
    "ThLoggerProcessor": ("logs", "ThLoggerProcessor"),
}


def __getattr__(name: str) -> Any:
    if name not in _EXPORT_MAP:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")
    module_name, attr_name = _EXPORT_MAP[name]
    module = import_module(f"{__name__}.{module_name}")
    value = getattr(module, attr_name)
    globals()[name] = value
    return value



# -*- coding: UTF-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 04.12.2023

Purpose: RB '/system/'
"""

from typing import Dict, Optional, Any

from ....logstool.queue import LoggerQueue

from ....attribtool import ReadOnlyClass
from ....logstool.logs import LoggerClient
from ..base import BRouterOS, BDev
from ...network.connectors import IConnector


class _Elements(object, metaclass=ReadOnlyClass):
    """Keys definition class.

    For internal purpose only.
    """

    ACTION: str = "action"
    BACKUP: str = "backup"
    CLIENT: str = "client"
    CLOCK: str = "clock"
    CLOUD: str = "cloud"
    CONSOLE: str = "console"
    CPU: str = "cpu"
    DEVICE_MODE: str = "device-mode"
    ENVIRONMENT: str = "environment"
    GAUGES: str = "gauges"
    HEALTH: str = "health"
    HISTORY: str = "history"
    IDENTITY: str = "identity"
    IRQ: str = "irq"
    JOB: str = "job"
    KEY: str = "key"
    LEDS: str = "leds"
    LICENSE: str = "license"
    LOGGING: str = "logging"
    MANUAL: str = "manual"
    NOTE: str = "note"
    NTP: str = "ntp"
    PACKAGE: str = "package"
    PCI: str = "pci"
    RESOURCE: str = "resource"
    ROOT: str = "system"
    ROUTERBOARD: str = "routerboard"
    SCHEDULER: str = "scheduler"
    SCRIPT: str = "script"
    SERVER: str = "server"
    SERVERS: str = "servers"
    SETTINGS: str = "settings"
    UPDATE: str = "update"
    UPGRADE: str = "upgrade"
    USB: str = "usb"
    WATCHDOG: str = "watchdog"


class RBSystem(BRouterOS):
    """System class

    For command root: /system/
    """

    def __init__(
        self,
        parent: BDev,
        connector: IConnector,
        qlog: Optional[LoggerQueue] = None,
        debug: bool = False,
        verbose: bool = False,
    ) -> None:
        """Constructor.

        ### Arguments:
        * parent: BDev - Parent device object.
        * connector: IConnector - Network connector interface for RouterOS communication.
        * qlog: Optional[LoggerQueue] - Logger queue for logging operations. Defaults to None.
        * debug: bool - Enable debug mode. Defaults to False.
        * verbose: bool - Enable verbose output. Defaults to False.
        """
        super().__init__(
            parent,
            connector,
            LoggerClient(queue=qlog, name=self._c_name),
            debug,
            verbose,
        )
        self.root = f"{_Elements.ROOT}/"

        # add elements
        elements: Dict[str, Any] = {
            _Elements.BACKUP: {
                _Elements.CLOUD: {},
            },
            _Elements.CLOCK: {
                _Elements.MANUAL: {},
            },
            _Elements.CONSOLE: {},
            _Elements.DEVICE_MODE: {},
            _Elements.HEALTH: {
                _Elements.SETTINGS: {},
                _Elements.GAUGES: {},
            },
            _Elements.HISTORY: {},
            _Elements.IDENTITY: {},
            _Elements.LEDS: {
                _Elements.SETTINGS: {},
            },
            _Elements.LICENSE: {},
            _Elements.LOGGING: {
                _Elements.ACTION: {},
            },
            _Elements.NOTE: {},
            _Elements.NTP: {
                _Elements.CLIENT: {
                    _Elements.SERVERS: {},
                },
                _Elements.KEY: {},
                _Elements.SERVER: {},
            },
            _Elements.PACKAGE: {
                _Elements.UPDATE: {},
            },
            _Elements.RESOURCE: {
                _Elements.CPU: {},
                _Elements.IRQ: {},
                _Elements.PCI: {},
                _Elements.USB: {
                    _Elements.SETTINGS: {},
                },
            },
            _Elements.ROUTERBOARD: {
                _Elements.SETTINGS: {},
            },
            _Elements.SCHEDULER: {},
            _Elements.SCRIPT: {
                _Elements.ENVIRONMENT: {},
                _Elements.JOB: {},
            },
            _Elements.UPGRADE: {},
            _Elements.WATCHDOG: {},
        }

        # configure elements
        self._add_elements(self, elements)


# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
Author:  Jacek Kotlarski --<szumak@virthost.pl>
Created: 23.06.2023

Purpose: Classes for IPv4
"""

import socket
import struct
from copy import deepcopy
from inspect import currentframe
import warnings
from typing import Iterator, Optional, TypeVar, Union, List

from ..attribtool import NoDynamicAttributes
from ..raisetool import Raise
from .libs.octets import Octet
from ..libs.interfaces.comparators import IComparators
from ..basetool.classes import BClasses

TAddress = TypeVar("TAddress", bound="Address")

DEFAULT_IPV4_HOST_LIMIT: int = 65536
DEFAULT_IPV4_SUBNET_LIMIT: int = 4096


class Address(IComparators, BClasses, NoDynamicAttributes):
    """Address class for representing IPv4 addresses.

    Constructor arguments:
    addr: Union[str, int, List[Octets]] -- IPv4 address representation as string, integer or list of four Octets

    Public property:
    octets: List[Octet] -- Return list of four Octets

    Public setter:
    octets: Union[str, int, List] -- Set IPv4 address from string, integer or list of octets.
    """

    __var_int: int = 0

    def __init__(
        self, addr: Union[str, int, Union[List[str], List[int], List[Octet]]]
    ) -> None:
        """Constructor.

        ### Arguments:
        * addr: Union[str, int, Union[List[str], List[int], List[Octet]]] - IPv4 address in various formats (string, integer, or list of octets).
        """
        self.octets = addr

    def __eq__(self, arg: Union[TAddress, object]) -> bool:
        """Equal."""
        if not isinstance(arg, Address):
            raise Raise.error(
                f"Expected argument of Address type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) == int(arg)

    def __ge__(self, arg: Union[TAddress, object]) -> bool:
        """Greater or equal."""
        if not isinstance(arg, Address):
            raise Raise.error(
                f"Expected argument of Address type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) >= int(arg)

    def __gt__(self, arg: Union[TAddress, object]) -> bool:
        """Greater."""
        if not isinstance(arg, Address):
            raise Raise.error(
                f"Expected argument of Address type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
            return False
        return int(self) > int(arg)

    def __le__(self, arg: Union[TAddress, object]) -> bool:
        """Less or equal."""
        if not isinstance(arg, Address):
            raise Raise.error(
                f"Expected argument of Address type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
        return int(self) <= int(arg)

    def __lt__(self, arg: Union[TAddress, object]) -> bool:
        """Less."""
        if not isinstance(arg, Address):
            raise Raise.error(
                f"Expected argument of Address type, received: {type(arg)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )
            return False
        return int(self) < int(arg)

    def __ne__(self, arg: Union[TAddress, object]) -> bool:
        """Negative."""
        if not isinstance(arg, Address):
            return False
        return int(self) != int(arg)

    @staticmethod
    def __int_to_ip(ip_int: int) -> str:
        """Convert ip int representation to ipv4 str."""
        return socket.inet_ntoa(struct.pack("!L", ip_int))

    @staticmethod
    def __ip_to_int(ip_str: str) -> int:
        """Convert ipv4 str representation to ip int."""
        return struct.unpack("!L", socket.inet_aton(ip_str))[0]

    def __set_octets_from_list(
        self, value: Union[List[str], List[int], List[Octet]]
    ) -> None:
        if not value:
            raise Raise.error(
                "Empty list received.",
                ValueError,
                self._c_name,
                currentframe(),
            )
        if len(value) != 4:
            raise Raise.error(
                f"Expected list with four elements, len({len(value)}) received.",
                ValueError,
                self._c_name,
                currentframe(),
            )

        self.__var_int = Address.__ip_to_int(
            f"{Octet(value[0])}.{Octet(value[1])}.{Octet(value[2])}.{Octet(value[3])}"
        )

    def __set_octets_from_int(self, value: int) -> None:
        if value in range(0, 4294967296):
            self.__var_int = value
        else:
            raise Raise.error(
                f"IP-int out of range (0-4294967295), received: {value}",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __set_octets_from_str(self, value: str) -> None:
        self.__set_octets_from_list(value.split("."))

    def __int__(self) -> int:
        """Return ipv4 representation as integer."""
        return self.__var_int

    def __str__(self) -> str:
        """Return string representation of address."""
        return Address.__int_to_ip(self.__var_int)

    def __repr__(self) -> str:
        """Return representation of object."""
        return f"{self._c_name}('{str(self)}')"

    @property
    def octets(self) -> List[Octet]:
        """Return octets list of four Octets.

        ### Returns:
        List[Octet] - List of four Octet objects representing each byte.
        """
        tmp: list[str] = str(self).split(".")
        return [Octet(tmp[0]), Octet(tmp[1]), Octet(tmp[2]), Octet(tmp[3])]

    @octets.setter
    def octets(
        self, value: Union[str, int, Union[List[str], List[int], List[Octet]]]
    ) -> None:
        """Set IPv4 address from string, integer or list of octets.

        ### Arguments:
        * value: Union[str, int, Union[List[str], List[int], List[Octet]]] - IPv4 address representation.
        """
        if isinstance(value, List):
            self.__set_octets_from_list(value)
        elif isinstance(value, int):
            self.__set_octets_from_int(value)
        elif isinstance(value, str):
            self.__set_octets_from_str(value)
        else:
            raise Raise.error(
                f"Expected String or Integer or List type, received: {type(value)}.",
                TypeError,
                self._c_name,
                currentframe(),
            )


# netmask
class Netmask(BClasses, NoDynamicAttributes):
    """Netmask class for IPv4 addresses.

    Constructor argument:
    addr: Union[str, int, List] -- Set netmask from string, integer or list of proper format of netmask octets.

    Public property:
    octets: List[Octet] -- Return netmask as list of four octets.
    cidr: str -- Return netmask in CIDR string format.

    Public setter:
    octets: List[Octet] -- Set netmask from list of 4 values [int||str||Octets].
    cidr: Union[str, int] -- Set netmask from CIDR format of string or integer.
    """

    # CIDR format
    __cidr: int = 0

    def __init__(
        self, addr: Union[str, int, Union[List[str], List[int], List[Octet]]]
    ) -> None:
        """Constructor.

        ### Arguments:
        * addr: Union[str, int, Union[List[str], List[int], List[Octet]]] - IPv4 netmask in various formats (string, integer, or list of octets).
        """
        if isinstance(addr, int):
            self.cidr = addr
        elif isinstance(addr, str):
            if len(addr) < 3:
                self.cidr = addr
            else:
                self.octets = addr
        elif isinstance(addr, List):
            self.octets = str(Address(addr))
        else:
            raise Raise.error(
                f"Expected String, Integer or List type, received: '{type(addr)}'.",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __int__(self) -> int:
        return self.__cidr

    def __str__(self) -> str:
        # convert CIDR to netmask
        return socket.inet_ntoa(
            struct.pack("!I", (1 << 32) - (1 << (32 - self.__cidr)))
        )

    def __repr__(self) -> str:
        return f"{self._c_name}({self.cidr})"

    def __cidr_validator(self, cidr: int) -> None:
        """Check and set cidr."""
        if cidr >= 0 and cidr <= 32:
            self.__cidr = cidr
        else:
            raise Raise.error(
                f"CIDR is out of range (0-32), received: {cidr}",
                ValueError,
                self._c_name,
                currentframe(),
            )

    @staticmethod
    def __octets_validator(octets: int) -> bool:
        """Check if given octets list is valid."""
        test: list[int] = [
            0,  # '0.0.0.0'
            2147483648,  # "128.0.0.0",
            3221225472,  # "192.0.0.0",
            3758096384,  # "224.0.0.0",
            4026531840,  # "240.0.0.0",
            4160749568,  # "248.0.0.0",
            4227858432,  # "252.0.0.0",
            4261412864,  # "254.0.0.0",
            4278190080,  # "255.0.0.0",
            4286578688,  # "255.128.0.0",
            4290772992,  # "255.192.0.0",
            4292870144,  # "255.224.0.0",
            4293918720,  # "255.240.0.0",
            4294443008,  # "255.248.0.0",
            4294705152,  # "255.252.0.0",
            4294836224,  # "255.254.0.0",
            4294901760,  # "255.255.0.0",
            4294934528,  # "255.255.128.0",
            4294950912,  # "255.255.192.0",
            4294959104,  # "255.255.224.0",
            4294963200,  # "255.255.240.0",
            4294965248,  # "255.255.248.0",
            4294966272,  # "255.255.252.0",
            4294966784,  # "255.255.254.0",
            4294967040,  # "255.255.255.0",
            4294967168,  # "255.255.255.128",
            4294967232,  # "255.255.255.192",
            4294967264,  # "255.255.255.224",
            4294967280,  # "255.255.255.240",
            4294967288,  # "255.255.255.248",
            4294967292,  # "255.255.255.252",
            4294967294,  # "255.255.255.254",
            4294967295,  # "255.255.255.255",
        ]
        if octets in test:
            return True
        return False

    @property
    def octets(self) -> List[Octet]:
        """Return octets list of four Octets.

        ### Returns:
        List[Octet] - List of four Octet objects representing each byte.
        """
        tmp: list[str] = str(self).split(".")
        return [Octet(tmp[0]), Octet(tmp[1]), Octet(tmp[2]), Octet(tmp[3])]

    @octets.setter
    def octets(
        self, addr: Union[str, int, Union[List[str], List[int], List[Octet]]]
    ) -> None:
        """Set netmask from string, integer or list of proper format of netmask octets.

        ### Arguments:
        * addr: Union[str, int, Union[List[str], List[int], List[Octet]]] - Netmask representation.
        """
        tmp = int(Address(addr))
        if not Netmask.__octets_validator(tmp):
            raise Raise.error(
                f"Invalid mask, received: {str(Address(addr))}",
                ValueError,
                self._c_name,
                currentframe(),
            )
        self.cidr = sum([bin(x.value).count("1") for x in Address(addr).octets])

    @property
    def cidr(self) -> str:
        """Return CIDR netmask as string type.

        ### Returns:
        str - CIDR notation string (e.g., '255.255.255.0').
        """
        return str(self.__cidr)

    @cidr.setter
    def cidr(self, value: Union[str, int]) -> None:
        """Set netmask from CIDR format of string or integer.

        ### Arguments:
        * value: Union[str, int] - CIDR notation (0-32).
        """
        if isinstance(value, str) and value.isdigit():
            self.__cidr_validator(int(value))
        elif isinstance(value, int):
            self.__cidr_validator(value)
        else:
            raise Raise.error(
                f"Expected Digit string or Int, received: {value}",
                ValueError,
                self._c_name,
                currentframe(),
            )


# Network
class Network(BClasses, NoDynamicAttributes):
    """Network IPv4 class.

    Constructor argument:
    addr: Union[str, List] -- Set IPv4 network address from string or two element list of address [Address,str,int,list] and netmask [Netmask, str, int, list].

    Public property:
    address: Address -- Return IPv4 address set in the constructor.
    broadcast: Address -- Return broadcast address.
    count: int -- Return count hosts addresses in network range.
    hosts(limit: Optional[int] = DEFAULT_IPV4_HOST_LIMIT): List[Address] -- Deprecated helper returning host list with an optional safety limit.
    iter_hosts(limit: Optional[int] = DEFAULT_IPV4_HOST_LIMIT) -> Iterator[Address] -- Lazy host iterator respecting optional limits.
    network: Address -- Return network address.
    mask: Netmask -- Return netmask.
    max: Address -- Return max address of host in network range.
    min: Address -- Return min address of host in network range.
    """

    __address: Address = None  # type: ignore
    __mask: Netmask = None  # type: ignore

    def __init__(self, addr: Union[str, List]) -> None:
        """Constructor.

        ### Arguments:
        * addr: Union[str, List] - IPv4 network address in CIDR notation (string) or list format.
        """
        if isinstance(addr, str):
            self.__network_from_str(addr)
        elif isinstance(addr, List):
            self.__network_from_list(addr)
        else:
            raise Raise.error(
                f"Expected IP network string or list type, received: '{type(addr)}'.",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __str__(self) -> str:
        """Return string representation of network address."""
        return f"{self.network}/{int(self.mask)}"

    def __repr__(self) -> str:
        """Return  string representation of class object."""
        return f"{self._c_name}({str(self)})"

    def __network_from_str(self, addr: str) -> None:
        """Build configuration from string."""
        if addr.find("/") > 0:
            tmp: list[str] = addr.split("/")
            self.__address = Address(tmp[0])
            self.__mask = Netmask(tmp[1])
        else:
            raise Raise.error(
                f"Expected network address in 'ip/mask' format string, received: '{addr}'.",
                ValueError,
                self._c_name,
                currentframe(),
            )

    def __network_from_list(self, addr: List) -> None:
        """Build configuration from list."""
        if len(addr) != 2:
            raise Raise.error(
                "Expected two element list ['ip','netmask']",
                ValueError,
                self._c_name,
                currentframe(),
            )
        if isinstance(addr[0], Address):
            self.__address = deepcopy(addr[0])
        else:
            self.__address = Address(addr[0])
        if isinstance(addr[1], Netmask):
            self.__mask = deepcopy(addr[1])
        else:
            self.__mask = Netmask(addr[1])

    @property
    def address(self) -> Address:
        """Return IPv4 address.

        ### Returns:
        Address - The IPv4 address object.
        """
        return self.__address

    @property
    def broadcast(self) -> Address:
        """Return broadcast address.

        ### Returns:
        Address - The broadcast address for this network.
        """
        ip = int(self.address)
        mask = int(Address(self.mask.octets))
        broadcast: int = ip | (mask ^ (1 << 32) - 1)
        return Address(broadcast)

    @property
    def count(self) -> int:
        """Return count hosts addresses in network range.

        ### Returns:
        int - Number of host addresses in the network.
        """
        net = int(self.network)
        broadcast = int(self.broadcast)
        return broadcast - net - 1 if broadcast - net > 2 else 0

    def hosts(self, limit: Optional[int] = DEFAULT_IPV4_HOST_LIMIT) -> List[Address]:
        """Return list of hosts in network range. (Deprecated)

        ### Arguments:
        * limit: Optional[int] - Maximum number of hosts allowed to materialise. Defaults to DEFAULT_IPV4_HOST_LIMIT.

        ### Returns:
        [List[Address]] - List of host addresses within the network.

        ### Raises:
        * ValueError: Propagated from `iter_hosts()` when the number of hosts exceeds the configured limit.
        """
        warnings.warn(
            "Network.hosts() is deprecated; use Network.iter_hosts() for lazy iteration.",
            DeprecationWarning,
            stacklevel=2,
        )
        return list(self.iter_hosts(limit=limit))

    def iter_hosts(
        self, limit: Optional[int] = DEFAULT_IPV4_HOST_LIMIT
    ) -> Iterator[Address]:
        """Yield hosts in network range lazily.

        ### Arguments:
        * limit: Optional[int] - Maximum number of hosts allowed before raising. Defaults to DEFAULT_IPV4_HOST_LIMIT.

        ### Returns:
        Iterator[Address] - Generator producing host addresses in ascending order.

        ### Raises:
        * ValueError: Raised when the number of hosts exceeds the configured limit.
        """
        effective_limit = limit
        host_count = self.count
        if effective_limit is not None and host_count > effective_limit:
            raise Raise.error(
                (
                    f"Network host count ({host_count}) exceeds limit "
                    f"({effective_limit}). Use iter_hosts(limit=None) for explicit override."
                ),
                ValueError,
                self._c_name,
                currentframe(),
            )
        net = int(self.network)
        broadcast = int(self.broadcast)
        for i in range(1, broadcast - net):
            yield Address(net + i)

    @property
    def mask(self) -> Netmask:
        """Return IPv4 network mask.

        ### Returns:
        Netmask - The network mask object.
        """
        return self.__mask

    @property
    def max(self) -> Address:
        """Return last address of host in network range.

        ### Returns:
        Address - The last usable host address.
        """
        net = int(self.network)
        broadcast = int(self.broadcast)
        ip = broadcast - 1
        return Address(ip) if ip > net else self.broadcast

    @property
    def min(self) -> Address:
        """Return first host address in network range.

        ### Returns:
        Address - The first usable host address.
        """
        net = int(self.network)
        broadcast = int(self.broadcast)
        ip: int = net + 1
        return Address(ip) if ip < broadcast else self.network

    @property
    def network(self) -> Address:
        """Return network address.

        ### Returns:
        Address - The network address.
        """
        ip = int(self.address)
        mask = int(Address(self.mask.octets))
        net: int = ip & mask
        return Address(net)


# SubNetwork
class SubNetwork(BClasses, NoDynamicAttributes):
    """SubNetwork calculator class.

    Constructor argument:
    network: Network -- The address of the network where the subnet is being searched for.
    mask: Netmask -- Subnet mask.

    Public property:
    subnets(limit: Optional[int] = DEFAULT_IPV4_SUBNET_LIMIT): List[Network] -- Deprecated helper returning subnet list with an optional safety limit.
    iter_subnets(limit: Optional[int] = DEFAULT_IPV4_SUBNET_LIMIT) -> Iterator[Network] -- Lazy subnet iterator.
    """

    __network: Network = None  # type: ignore
    __mask: Netmask = None  # type: ignore

    def __init__(self, network: Network, mask: Netmask) -> None:
        """Constructor.

        ### Arguments:
        * network: Network - Network address object.
        * mask: Netmask - Netmask object defining the subnet size.
        """
        if isinstance(network, Network) and isinstance(mask, Netmask):
            if int(network.mask) <= int(mask):
                self.__network = network
                self.__mask = mask
            else:
                raise Raise.error(
                    (
                        "The network mask must be greater then or equal to the subnet mask you are looking for."
                        f"Received: {int(network.mask)} and {int(mask)}"
                    ),
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
        else:
            raise Raise.error(
                f"Expected argument of (Network, Netmask), received: ({type(network)},{type(mask)}).",
                TypeError,
                self._c_name,
                currentframe(),
            )

    def subnets(
        self, limit: Optional[int] = DEFAULT_IPV4_SUBNET_LIMIT
    ) -> List[Network]:
        """Return subnets list. (Deprecated)

        ### Arguments:
        * limit: Optional[int] - Maximum number of subnets to generate. Defaults to DEFAULT_IPV4_SUBNET_LIMIT.

        ### Returns:
        [List[Network]] - List of generated subnetworks.

        ### Raises:
        * ValueError: Propagated from `iter_subnets()` when the number of subnets exceeds the configured limit.
        """
        warnings.warn(
            "SubNetwork.subnets() is deprecated; use SubNetwork.iter_subnets() for lazy iteration.",
            DeprecationWarning,
            stacklevel=2,
        )
        return list(self.iter_subnets(limit=limit))

    def iter_subnets(
        self, limit: Optional[int] = DEFAULT_IPV4_SUBNET_LIMIT
    ) -> Iterator[Network]:
        """Yield IPv4 subnetworks lazily.

        ### Arguments:
        * limit: Optional[int] - Maximum number of subnetworks allowed before raising. Defaults to DEFAULT_IPV4_SUBNET_LIMIT.

        ### Returns:
        Iterator[Network] - Generator producing IPv4 subnetworks in ascending order.

        ### Raises:
        * ValueError: Raised when the number of subnetworks exceeds the configured limit.
        """
        effective_limit = limit
        produced = 0
        net_start = int(self.__network.network)
        net_end = int(self.__network.broadcast)
        start: int = net_start
        while True:
            if effective_limit is not None and produced >= effective_limit:
                raise Raise.error(
                    (
                        f"Subnet count exceeds limit ({effective_limit}). "
                        "Use iter_subnets(limit=None) for explicit override."
                    ),
                    ValueError,
                    self._c_name,
                    currentframe(),
                )
            subnet = Network([Address(start), self.__mask])
            yield subnet
            produced += 1
            if int(subnet.broadcast) >= net_end:
                break
            start = int(subnet.broadcast) + 1


# #[EOF]#######################################################################



# -*- coding: utf-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 2024-01-15

Purpose: Provide a typed container mixin with helper accessors.

`BData` ensures subclasses can manage internal dictionaries with optional type
constraints, reliable copying, and lifecycle utilities for managed state.
"""

import copy
import warnings
import sys

from inspect import currentframe
from typing import Dict, List, Any, Optional, Type, Union, get_origin, get_args

from ..raisetool import Raise

from .classes import BClasses


class BData(BClasses):
    """Base mixin that adds typed dictionary semantics."""

    __bdata_storage: Optional[Dict[str, Any]] = None
    __bdata_types: Optional[Dict[str, Any]] = None

    @staticmethod
    def __validate_type(value: Any, expected_type: Any) -> bool:
        """Validate if value matches expected type including complex generic types.

        ### Arguments:
        * value: Any - Value to check.
        * expected_type: Any - Expected type (can be simple or generic type from typing).

        ### Returns:
        [bool] - True if value matches expected type.

        ### Note:
        Supports simple types (int, str, list) and complex generic types
        (Optional[str], Dict[str, int], List[Any], etc.).
        """
        # Handle None for Optional types
        if value is None:
            origin = get_origin(expected_type)
            # Check if it's Optional (Union with None)
            if origin is Union:
                args = get_args(expected_type)
                return type(None) in args
            # None is only valid for Optional types
            return False

        # Try simple isinstance first (works for simple types and some generics)
        try:
            if isinstance(value, expected_type):
                return True
        except TypeError:
            # isinstance() doesn't work with some generic types, continue to detailed check
            pass

        # Get origin and args for generic types
        origin = get_origin(expected_type)
        args = get_args(expected_type)

        # No origin means it's a simple type - use isinstance
        if origin is None:
            try:
                return isinstance(value, expected_type)
            except TypeError:
                return False

        # Handle Union types (including Optional)
        if origin is Union:
            # Check if value matches any of the union types
            return any(BData.__validate_type(value, arg) for arg in args)

        # Handle List, list
        if origin in (list, List):
            if not isinstance(value, list):
                return False
            # If no type args or List[Any], any list is valid
            if not args or args == (Any,):
                return True
            # Check all elements match the specified type
            element_type = args[0]
            return all(BData.__validate_type(item, element_type) for item in value)

        # Handle Dict, dict
        if origin in (dict, Dict):
            if not isinstance(value, dict):
                return False
            # If no type args or Dict[Any, Any], any dict is valid
            if not args or len(args) < 2:
                return True
            key_type, value_type = args[0], args[1]
            # Check all keys and values match specified types
            return all(
                BData.__validate_type(k, key_type)
                and BData.__validate_type(v, value_type)
                for k, v in value.items()
            )

        # Handle Tuple
        if origin is tuple:
            if not isinstance(value, tuple):
                return False
            if not args:
                return True
            # Variable length tuple: Tuple[int, ...]
            if len(args) == 2 and args[1] is Ellipsis:
                element_type = args[0]
                return all(BData.__validate_type(item, element_type) for item in value)
            # Fixed length tuple: Tuple[int, str, bool]
            if len(value) != len(args):
                return False
            return all(BData.__validate_type(v, t) for v, t in zip(value, args))

        # Handle Set, set
        if origin in (set, frozenset):
            if not isinstance(value, (set, frozenset)):
                return False
            if not args or args == (Any,):
                return True
            element_type = args[0]
            return all(BData.__validate_type(item, element_type) for item in value)

        # For other generic types, fall back to origin check
        try:
            return isinstance(value, origin)
        except TypeError:
            return False

    def __check_keys(self, key: str) -> bool:
        """Check if the key is available in the storage dictionary.

        ### Arguments:
        * key: str - Dictionary key to verify.

        ### Returns:
        [bool] - True when the key exists, False otherwise.
        """
        if self.__bdata_storage and key in self.__bdata_storage:
            return True
        return False

    def __has_type(self, key: str) -> bool:
        """Check if a type hint is registered for the given key.

        ### Arguments:
        * key: str - Dictionary key to verify.

        ### Returns:
        [bool] - True when a type constraint is registered.
        """
        if self.__bdata_types and key in self.__bdata_types:
            return True
        return False

    def __check_type(self, key: str, received_type: Optional[Any]) -> bool:
        """Validate that the stored type matches the received one.

        ### Arguments:
        * key: str - Dictionary key whose type should be verified.
        * received_type: Optional[Any] - Type extracted from input data.

        ### Returns:
        [bool] - True when the type matches the stored constraint.

        ### Raises:
        * KeyError: Raised when the key has no registered type constraint.
        """
        if self.__bdata_types and self.__has_type(key):
            if received_type == self.__bdata_types[key]:
                return True
            return False
        raise Raise.error(
            f'The key: "{key}" is missing in the directory.',
            KeyError,
            self._c_name,
            currentframe(),
        )

    def _copy_data(self, key: str) -> Optional[Any]:
        """Copy data from the internal dictionary.

        ### Arguments:
        * key: str - Variable name to copy.

        ### Returns:
        [Optional[Any]] - Deep copy of the stored value or None when missing.
        """
        if self.__check_keys(key):
            return copy.deepcopy(self._data[key])
        return None

    def _get_data(
        self,
        key: str,
        default_value: Optional[Any] = None,
        **kwargs,
    ):
        """Return data from the internal dictionary.

        ### Arguments:
        * key: str - Variable name.
        * default_value: Optional[Any] - Fallback value when the key is missing.

        ### Returns:
        [Optional[Any]] - Stored value or provided default.

        ### Raises:
        * TypeError: Default value does not match the registered type.

        ### Note:
        The `set_default_type` parameter is deprecated and ignored if passed.
        Use `_set_data()` with `set_default_type` to register type constraints.
        """
        # Handle deprecated set_default_type parameter for backward compatibility
        if "set_default_type" in kwargs:
            warnings.warn(
                "The 'set_default_type' parameter in _get_data() is deprecated. "
                "Use _set_data() with set_default_type to register type constraints.",
                DeprecationWarning,
                stacklevel=2,
            )

        if self.__check_keys(key):
            return self._data[key]

        if default_value is not None:
            if (
                self.__bdata_types
                and self.__has_type(key)
                and not self.__validate_type(default_value, self.__bdata_types[key])
            ):
                raise Raise.error(
                    f"Expected '{self.__bdata_types[key]}' type, received default_value type is: {type(default_value)}",
                    TypeError,
                    self._c_name,
                    currentframe(),
                )
            return default_value
        return None

    def _set_data(
        self,
        key: str,
        value: Optional[Any],
        set_default_type: Optional[Any] = None,
    ) -> None:
        """Set data in the internal dictionary.

        ### Arguments:
        * key: str - Variable name.
        * value: Optional[Any] - Value to assign.
        * set_default_type: Optional[Any] - Optional type restriction for the key.
          If None, no type constraint is registered (existing constraints are preserved).

        ### Raises:
        * TypeError: Value violates the registered or provided type constraint.
        * TypeError: Attempting to overwrite an existing type constraint.
          Use _delete_data(key) first to change the type.
        * TypeError: default_value type does not match set_default_type.
        """
        if self.__bdata_types is None:
            self.__bdata_types = {}

        if self.__has_type(key):
            # Key already has a registered type
            if (
                set_default_type is not None
                and set_default_type != self.__bdata_types[key]
            ):
                # Attempting to change the type without deleting first
                raise Raise.error(
                    f"Cannot overwrite existing type constraint for key '{key}'. "
                    f"Current type: '{self.__bdata_types[key]}', attempted type: '{set_default_type}'. "
                    f"Use _delete_data('{key}') first to change the type constraint.",
                    TypeError,
                    self._c_name,
                    currentframe(),
                )

            # Verify value matches existing type
            if self.__validate_type(value, self.__bdata_types[key]):
                self._clear_data(key)
                self._data[key] = value
            else:
                raise Raise.error(
                    f"Expected '{self.__bdata_types[key]}' type, received: '{type(value)}'",
                    TypeError,
                    self._c_name,
                    currentframe(),
                )
        else:
            # No type registered yet
            if set_default_type is not None:
                # Register new type and verify value matches
                self.__bdata_types[key] = set_default_type
                if self.__validate_type(value, set_default_type):
                    self._data[key] = value
                else:
                    # Clean up type registration if value doesn't match
                    del self.__bdata_types[key]
                    raise Raise.error(
                        f"The type of the value: '{type(value)}' does not match the type passed in the 'set_default_type': '{set_default_type}' variable",
                        TypeError,
                        self._c_name,
                        currentframe(),
                    )
            else:
                # No type constraint - accept any type
                self._clear_data(key)
                self._data[key] = value

    def _delete_data(self, key: str) -> None:
        """Delete data and type constraint from the internal dictionary.

        ### Arguments:
        * key: str - Variable name to delete.
        """
        if self.__check_keys(key):
            del self._data[key]
        if self.__has_type(key):
            del self.__bdata_types[key]  # type: ignore

    def _clear_data(self, key: str) -> None:
        """Clear data from the internal dictionary.

        Preserve the key type constraint. If the key is missing, do nothing.

        ### Arguments:
        * key: str - Variable name to delete.
        """
        if self.__check_keys(key):
            if isinstance(self._data[key], (List, Dict)):
                self._data[key].clear()
            del self._data[key]

    @property
    def _data(self) -> Dict[str, Any]:
        """Return the internal data dictionary, initializing if needed.

        ### Returns:
        [Dict[str, Any]] - Mutable storage dictionary.
        """
        if self.__bdata_storage is None:
            self.__bdata_storage = {}
        if self.__bdata_types is None:
            self.__bdata_types = {}
        return self.__bdata_storage

    @_data.setter
    def _data(self, value: Optional[Dict[str, Any]]) -> None:
        """Assign a dictionary to the internal storage.

        ### Arguments:
        * value: Optional[Dict[str, Any]] - New dictionary or None to clear.

        ### Raises:
        * TypeError: Provided value is neither None nor a dictionary.
        * TypeError: Incoming values conflict with registered type constraints.
        """
        if value is None:
            if self.__bdata_storage is not None:
                self.__bdata_storage.clear()
            if self.__bdata_types is not None:
                self.__bdata_types.clear()
            return None
        if isinstance(value, Dict) and self.__bdata_storage is not None:
            for key in value.keys():
                if self.__bdata_types and self.__has_type(key):
                    if not self.__check_type(key, type(value[key])):
                        raise Raise.error(
                            f"Expected '{self.__bdata_types[key]}' type, received: '{type(value[key])}'",
                            TypeError,
                            self._c_name,
                            currentframe(),
                        )
                self.__bdata_storage[key] = value[key]
        else:
            raise Raise.error(
                f"Expected Dict type, received: '{type(value)}'.",
                TypeError,
                self._c_name,
                currentframe(),
            )

    @_data.deleter
    def _data(self) -> None:
        """Delete the data dictionary and registered types."""
        if self.__bdata_storage is not None:
            self.__bdata_storage.clear()
        if self.__bdata_types is not None:
            self.__bdata_types.clear()
        self.__bdata_storage = None
        self.__bdata_types = None


#
# #[EOF]#######################################################################



# -*- coding: UTF-8 -*-
"""
Author:  Jacek Kotlarski --<szumak@virthost.pl>
Created: 2024-01-15

Purpose: Provide Tkinter toolkit exports with lazy loading.

The module exposes base mixins, geometry helpers, and composite widgets on demand so that projects
can import the package without immediately initialising Tkinter or loading heavy helpers.
"""

from importlib import import_module
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from .base import TkBase as TkBase
    from .layout import Grid as Grid
    from .layout import Pack as Pack
    from .layout import Place as Place
    from .tools import ClipBoard as ClipBoard
    from .widgets import CreateToolTip as CreateToolTip
    from .widgets import StatusBarTkFrame as StatusBarTkFrame
    from .widgets import StatusBarTtkFrame as StatusBarTtkFrame
    from .widgets import VerticalScrolledTkFrame as VerticalScrolledTkFrame
    from .widgets import VerticalScrolledTtkFrame as VerticalScrolledTtkFrame

__all__ = [
    "TkBase",
    "Pack",
    "Grid",
    "Place",
    "ClipBoard",
    "StatusBarTkFrame",
    "StatusBarTtkFrame",
    "CreateToolTip",
    "VerticalScrolledTkFrame",
    "VerticalScrolledTtkFrame",
]

_EXPORT_MAP = {
    "TkBase": ("base", "TkBase"),
    "Pack": ("layout", "Pack"),
    "Grid": ("layout", "Grid"),
    "Place": ("layout", "Place"),
    "ClipBoard": ("tools", "ClipBoard"),
    "StatusBarTkFrame": ("widgets", "StatusBarTkFrame"),
    "StatusBarTtkFrame": ("widgets", "StatusBarTtkFrame"),
    "CreateToolTip": ("widgets", "CreateToolTip"),
    "VerticalScrolledTkFrame": ("widgets", "VerticalScrolledTkFrame"),
    "VerticalScrolledTtkFrame": ("widgets", "VerticalScrolledTtkFrame"),
}


def __getattr__(name: str) -> Any:
    """Resolve configured Tk toolkit exports on demand.

    ### Arguments:
    * name: str - Requested attribute name.

    ### Returns:
    [Any] - Resolved attribute from the target submodule.

    ### Raises:
    * AttributeError: Raised when the attribute is not registered.
    """
    if name not in _EXPORT_MAP:
        raise AttributeError(f"module '{__name__}' has no attribute '{name}'")
    module_name, attr_name = _EXPORT_MAP[name]
    module = import_module(f"{__name__}.{module_name}")
    value = getattr(module, attr_name)
    globals()[name] = value
    return value


def __dir__() -> list[str]:
    """Expose lazy exports to dir()."""
    return sorted(__all__)



# -*- coding: utf-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 2024-01-15

Purpose: Provide named accessors for Tk geometry manager constants (pack, grid, place).

This module centralises the most common anchor, side, fill, and sticky options so that GUI code can
refer to Tk enumerations through read-only wrappers instead of raw tkinter symbols.
"""

import tkinter as tk

from ..attribtool import ReadOnlyClass


class Pack(object, metaclass=ReadOnlyClass):
    """Pack geometry manager constants.

    Groups anchors, sides, and fill modes exposed by ``tk.pack`` for consistent, read-only access.
    """

    class Anchor(object, metaclass=ReadOnlyClass):
        """Anchor values for pack placement.

        Mirrors the ``anchor`` parameter accepted by the pack geometry manager.
        """

        CENTER = tk.CENTER
        E = tk.E
        N = tk.N
        NE = tk.NE
        NW = tk.NW
        S = tk.S
        SE = tk.SE
        SW = tk.SW
        W = tk.W

    class Side(object, metaclass=ReadOnlyClass):
        """Side values for pack placement.

        Indicates which edge of the container a widget should hug when packed.
        """

        BOTTOM = tk.BOTTOM
        LEFT = tk.LEFT
        RIGHT = tk.RIGHT
        TOP = tk.TOP

    class Fill(object, metaclass=ReadOnlyClass):
        """Fill modes for pack placement.

        Controls how a widget expands along the horizontal and vertical axes.
        """

        BOTH = tk.BOTH
        NONE = tk.NONE
        X = tk.X
        Y = tk.Y


class Grid(object, metaclass=ReadOnlyClass):
    """Grid geometry manager constants.

    Aggregates sticky options from ``tk.grid`` to keep UI layout code expressive.
    """

    class Sticky(object, metaclass=ReadOnlyClass):
        """Sticky values for grid placement.

        Specifies which cell edges a widget should adhere to when the cell grows.
        """

        CENTER = tk.CENTER
        E = tk.E
        N = tk.N
        NE = tk.NE
        NW = tk.NW
        S = tk.S
        SE = tk.SE
        SW = tk.SW
        W = tk.W


class Place(object, metaclass=ReadOnlyClass):
    """Place geometry manager constants.

    Wraps commonly used ``tk.place`` anchor options inside a read-only namespace.
    """

    class Anchor(object, metaclass=ReadOnlyClass):
        """Anchor values for absolute placement.

        Describes which widget point maps to the x/y coordinates supplied to ``place``.
        """

        CENTER = tk.CENTER
        E = tk.E
        N = tk.N
        NE = tk.NE
        NW = tk.NW
        S = tk.S
        SE = tk.SE
        SW = tk.SW
        W = tk.W


# #[EOF]#######################################################################



"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 2024-05-12

Purpose: Collect connector interfaces and implementations used by device
packages.

Importing from this namespace offers a stable entry point for consumers that
need to build or configure network adapters such as the RouterOS API connector.
"""

# #[EOF]#######################################################################



# -*- coding: utf-8 -*-
"""
Author:  Jacek 'Szumak' Kotlarski --<szumak@virthost.pl>
Created: 2024-10-07

Purpose: EDMC plugins individual logging subsystem classes.
"""

from inspect import currentframe
import logging
import os
from typing import Union, Optional, List, Dict
from logging.handlers import RotatingFileHandler
from queue import Queue, SimpleQueue

from ..edmctool.system import EnvLocal

from ..attribtool import ReadOnlyClass, NoDynamicAttributes
from ..basetool.data import BData
from ..raisetool import Raise


class _Keys(object, metaclass=ReadOnlyClass):
    """Internal Keys container class."""

    LOG_DATA: str = "__logger_data__"
    LOG_LEVEL: str = "__logger_level__"
    LOG_QUEUE: str = "__logger_queue__"
    LP_ENGINE: str = "__log_processor_engine__"
    LP_NAME: str = "__log_processor_name__"
    LP_LOGS_DIR: str = "__log_processor_logs_dir__"
    LP_MAX_BYTES: str = "__log_processor_max_bytes__"
    LP_BACKUP_COUNT: str = "__log_processor_backup_count__"


class Log(BData):
    """Create Log container class."""

    def __init__(self, level: int) -> None:
        """Class constructor.

        ### Arguments:
        * level: int - Logging level for this log instance.
        """
        # init data list
        self._set_data(key=_Keys.LOG_DATA, value=[], set_default_type=List)

        # init default loglevel
        ll_test = LogLevels()
        self._set_data(key=_Keys.LOG_LEVEL, value=ll_test.debug, set_default_type=int)

        if isinstance(level, int) and ll_test.has_key(level):
            self._set_data(key=_Keys.LOG_LEVEL, value=level)
        else:
            raise Raise.error(
                f"Int type level expected, '{type(level)}' received.",
                TypeError,
                self._c_name,
                currentframe(),
            )

    @property
    def loglevel(self) -> int:
        """Return loglevel.

        ### Returns:
        int - The configured log level.
        """
        return self._get_data(key=_Keys.LOG_LEVEL)  # type: ignore

    @property
    def log(self) -> List[str]:
        """Get list of logs.

        ### Returns:
        List[str] - List of log messages.
        """
        return self._get_data(
            key=_Keys.LOG_DATA,
        )  # type: ignore

    @log.setter
    def log(self, arg: Optional[Union[List, str]]) -> None:
        """Set data log.

        ### Arguments:
        * arg: Optional[Union[List, str]] - Log message or list of messages to append, None to clear.
        """
        if arg is None or (isinstance(arg, List) and not bool(arg)):
            # cleanup list of logs
            self._set_data(
                key=_Keys.LOG_DATA,
                value=[],
            )
        if isinstance(arg, List):
            for msg in arg:
                self.log.append(f"{msg}")
        elif arg is None:
            pass
        else:
            self.log.append(f"{arg}")


class LogProcessor(BData):
    """Log processor access API."""

    def __init__(
        self,
        name: str,
        logs_dir: Optional[str] = None,
        max_bytes: int = 100000,
        backup_count: int = 5,
    ) -> None:
        """Create instance class object for processing single message.

        ### Arguments:
        * name: str - Name of the application or logger.
        * logs_dir: Optional[str] - Directory path for log files. Defaults to tmpdir if not provided.
        * max_bytes: int - Maximum size of log file in bytes before rotation. Defaults to 100000.
        * backup_count: int - Number of backup log files to keep. Defaults to 5.
        """
        # name of app
        self._set_data(key=_Keys.LP_NAME, value=name, set_default_type=str)
        # logs directory
        if logs_dir is None:
            logs_dir = EnvLocal().tmpdir
        self._set_data(
            key=_Keys.LP_LOGS_DIR,
            value=logs_dir,
            set_default_type=str,
        )
        # max bytes
        self._set_data(
            key=_Keys.LP_MAX_BYTES,
            value=max_bytes,
            set_default_type=int,
        )
        # backup count
        self._set_data(
            key=_Keys.LP_BACKUP_COUNT,
            value=backup_count,
            set_default_type=int,
        )
        self.loglevel = LogLevels().notset
        self.__logger_init()

    def __del__(self) -> None:
        """Destroy log instance."""
        self.close()

    @property
    def __engine(self) -> logging.Logger:
        """Return logger instance.

        ### Returns:
        logging.Logger - The underlying Python logging engine.
        """
        return self._get_data(key=_Keys.LP_ENGINE)  # type: ignore

    @__engine.setter
    def __engine(self, arg: logging.Logger) -> None:
        """Sets engine instance.

        ### Arguments:
        * arg: logging.Logger - Logger engine instance to assign.
        """
        self._set_data(key=_Keys.LP_ENGINE, value=arg, set_default_type=logging.Logger)

    def __logger_init(self) -> None:
        """Initialize logger engine."""
        self.close()

        self.__engine = logging.getLogger(self._get_data(key=_Keys.LP_NAME))
        self.__engine.setLevel(LogLevels().debug)

        log_handler = RotatingFileHandler(
            filename=os.path.join(
                f"{self._get_data(key=_Keys.LP_LOGS_DIR)}",
                f"{self._get_data(key=_Keys.LP_NAME)}.log",
            ),
            maxBytes=self._get_data(key=_Keys.LP_MAX_BYTES),  # type: ignore
            backupCount=self._get_data(key=_Keys.LP_BACKUP_COUNT),  # type: ignore
        )

        log_handler.setLevel(self.loglevel)
        log_handler.setFormatter(logging.Formatter("%(asctime)s %(message)s"))
        self.__engine.addHandler(log_handler)
        self.__engine.info("Logger initialization complete")

    def close(self) -> None:
        """Close log subsystem."""
        if self.__engine is not None:
            for handler in self.__engine.handlers:
                handler.close()
                self.__engine.removeHandler(handler)

    def send(self, message: Log) -> None:
        """Send single message to log engine.

        ### Arguments:
        * message: Log - Log object containing messages and log level.

        ### Raises:
        * TypeError: Provided argument is not a Log instance.
        """
        if self.__engine is None:
            return
        lgl = LogLevels()
        if isinstance(message, Log):
            if message.loglevel == lgl.critical:
                for msg in message.log:
                    self.__engine.critical("%s", msg)
            elif message.loglevel == lgl.debug:
                for msg in message.log:
                    self.__engine.debug("%s", msg)
            elif message.loglevel == lgl.error:
                for msg in message.log:
                    self.__engine.error("%s", msg)
            elif message.loglevel == lgl.info:
                for msg in message.log:
                    self.__engine.info("%s", msg)
            elif message.loglevel == lgl.warning:
                for msg in message.log:
                    self.__engine.warning("%s", msg)
            else:
                for msg in message.log:
                    self.__engine.notset("%s", msg)  # type: ignore
        else:
            raise Raise.error(
                f"Log type expected, {type(message)} received.",
                TypeError,
                self._c_name,
                currentframe(),
            )

    @property
    def loglevel(self) -> int:
        """Property that returns loglevel.

        ### Returns:
        int - The configured log level.
        """
        return self._get_data(
            key=_Keys.LOG_LEVEL, default_value=LogLevels().notset
        )  # type: ignore

    @loglevel.setter
    def loglevel(self, arg: int) -> None:
        """Setter for log level parameter.

        ### Arguments:
        * arg: int - Log level constant from LogLevels.
        """
        if self.loglevel == arg:
            log = Log(LogLevels().debug)
            log.log = "LogLevel has not changed"
            self.send(log)
            return
        ll_test = LogLevels()
        if isinstance(arg, int) and ll_test.has_key(arg):
            self._set_data(key=_Keys.LOG_LEVEL, value=arg, set_default_type=int)
        else:
            tmp = "Unable to set LogLevel to {}, defaulted to INFO"
            log = Log(LogLevels().warning)
            log.log = tmp.format(arg)
            self.send(log)
            self._set_data(
                key=_Keys.LOG_LEVEL, value=LogLevels().info, set_default_type=int
            )
        self.__logger_init()


class LogClient(BData):
    """Log client class API."""

    def __init__(self, queue: Union[Queue, SimpleQueue]) -> None:
        """Create instance class object.

        ### Arguments:
        * queue: Union[Queue, SimpleQueue] - Queue object for sending log messages.
        """
        self._set_data(
            key=_Keys.LOG_QUEUE,
            value=queue,
            set_default_type=Union[Queue, SimpleQueue],
        )

    @property
    def queue(self) -> Union[Queue, SimpleQueue]:
        """Give me queue object.

        ### Returns:
        Union[Queue, SimpleQueue] - The queue object for log messages.
        """
        return self._get_data(
            key=_Keys.LOG_QUEUE,
        )  # type: ignore

    @property
    def critical(self) -> str:
        """Property that returns nothing.

        ### Returns:
        str - Empty string.
        """
        return ""

    @critical.setter
    def critical(self, message: Union[str, List]) -> None:
        """Setter for critical messages.

        ### Arguments:
        * message: Union[str, List] - Critical level log message or list of messages.
        """
        log = Log(LogLevels().critical)
        log.log = message
        self.queue.put(log)

    @property
    def debug(self) -> str:
        """Property that returns nothing.

        ### Returns:
        str - Empty string.
        """
        return ""

    @debug.setter
    def debug(self, message: Union[str, List]) -> None:
        """Setter for debug messages.

        ### Arguments:
        * message: Union[str, List] - Debug level log message or list of messages.
        """
        log = Log(LogLevels().debug)
        log.log = message
        self.queue.put(log)

    @property
    def error(self) -> str:
        """Property that returns nothing.

        ### Returns:
        str - Empty string.
        """
        return ""

    @error.setter
    def error(self, message: Union[str, List]) -> None:
        """Setter for error messages.

        ### Arguments:
        * message: Union[str, List] - Error level log message or list of messages.
        """
        log = Log(LogLevels().error)
        log.log = message
        self.queue.put(log)

    @property
    def info(self) -> str:
        """Property that returns nothing.

        ### Returns:
        str - Empty string.
        """
        return ""

    @info.setter
    def info(self, message: Union[str, List]) -> None:
        """Setter for info messages.

        ### Arguments:
        * message: Union[str, List] - Info level log message or list of messages.
        """
        log = Log(LogLevels().info)
        log.log = message
        self.queue.put(log)

    @property
    def warning(self) -> str:
        """Property that returns nothing.

        ### Returns:
        str - Empty string.
        """
        return ""

    @warning.setter
    def warning(self, message: Union[str, List]) -> None:
        """Setter for warning messages.

        ### Arguments:
        * message: Union[str, List] - Warning level log message or list of messages.
        """
        log = Log(LogLevels().warning)
        log.log = message
        self.queue.put(log)

    @property
    def notset(self) -> str:
        """Property that returns nothing.

        ### Returns:
        str - Empty string.
        """
        return ""

    @notset.setter
    def notset(self, message: Union[str, List]) -> None:
        """Setter for notset level messages.

        ### Arguments:
        * message: Union[str, List] - Message or list of messages to log.
        """
        log = Log(LogLevels().notset)
        log.log = message
        self.queue.put(log)


class LogLevels(NoDynamicAttributes):
    """Log levels keys.

    This is a container class with properties that return the proper
    logging levels defined in the logging module.
    """

    __keys: Dict[int, bool] = None  # type: ignore
    __txt: Dict[str, int] = None  # type: ignore

    def __init__(self) -> None:
        """Create Log instance."""
        # loglevel initialization database
        self.__keys = {
            self.info: True,
            self.debug: True,
            self.warning: True,
            self.error: True,
            self.notset: True,
            self.critical: True,
        }
        self.__txt = {
            "INFO": self.info,
            "DEBUG": self.debug,
            "WARNING": self.warning,
            "ERROR": self.error,
            "CRITICAL": self.critical,
            "NOTSET": self.notset,
        }

    def get(self, level: Union[int, str]) -> Optional[int]:
        """Get int log level."""
        if level in self.__txt:
            return self.__txt[level]
        return None

    def has_key(self, level: Union[int, str]) -> bool:
        """Check, if level is in proper keys.

        ### Returns:
        bool - True if level is valid, False otherwise.
        """
        if level in self.__keys or level in self.__txt:
            return True
        return False

    @property
    def info(self) -> int:
        """Return info level.

        ### Returns:
        int - The INFO log level constant.
        """
        return logging.INFO

    @property
    def debug(self) -> int:
        """Return debug level.

        ### Returns:
        int - The DEBUG log level constant.
        """
        return logging.DEBUG

    @property
    def warning(self) -> int:
        """Return warning level.

        ### Returns:
        int - The WARNING log level constant.
        """
        return logging.WARNING

    @property
    def error(self) -> int:
        """Return error level.

        ### Returns:
        int - The ERROR log level constant.
        """
        return logging.ERROR

    @property
    def critical(self) -> int:
        """Return critical level.

        ### Returns:
        int - The CRITICAL log level constant.
        """
        return logging.CRITICAL

    @property
    def notset(self) -> int:
        """Return notset level.

        ### Returns:
        int - The NOTSET log level constant.
        """
        return logging.NOTSET


# #[EOF]#######################################################################
